from typing import Callable, Tuple, Type, Union, overload

import numpy as np

from .. import structures as struct
from ..core import header as hdr

__all__ = [
    "all",
    "any",
    "argmax",
    "argmin",
    "max",
    "mean",
    "median",
    "min",
    "moment",
    "percentile",
    "ptp",
    "quantile",
    "std",
    "sum",
    "var",
    "rms",
]


def _reduce(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]],
    operator: Callable,
) -> Union["struct.Map", "struct.Profile", float]:
    """
    Convenience function to implement reducers.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if not isinstance(input, struct.Cube) and output_type is not float:
        raise TypeError(
            f"input type {type(input)} is not compatible with output type {output_type}"
        )
    if output_type is float:
        return operator(input.data)
    if isinstance(input, struct.Cube) and output_type is struct.Map:
        new_header = hdr.remove_header_axis(input.header, axis="spectral")
        new_data = operator(input.data, axis=0)
    elif isinstance(input, struct.Cube) and output_type is struct.Profile:
        new_header = hdr.remove_header_axis(input.header, axis="spatial")
        new_data = operator(input.data, axis=(1, 2))
    else:
        raise ValueError("Should never been here")
    return output_type(new_data, new_header)


def _arg(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]],
    operator: Callable,
) -> Union["struct.Map", "struct.Profile", tuple, int]:
    """
    Convenience function to implement argmin and argmax functions.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if not isinstance(input, struct.Cube) and output_type is not float:
        raise TypeError(
            f"input type {type(input)} is not compatible with output type {output_type}"
        )
    if output_type is float:
        index = operator(input.data)
        indices = np.unravel_index(index, input.data.shape)[::-1]
        if isinstance(input, struct.Profile):
            return indices[0]  # z
        return indices  # (x,y) or (x,y,z)
    if isinstance(input, struct.Cube) and output_type is struct.Map:
        new_header = hdr.remove_header_axis(input.header, axis="spectral")
        indices = operator(input.data, axis=0)
        indices = np.unravel_index(index, input.shape)
    elif isinstance(input, struct.Cube) and output_type is struct.Profile:
        new_header = hdr.remove_header_axis(input.header, axis="spatial")
        indices = operator(input.data, axis=(1, 2))
    else:
        raise ValueError("Should never been here")
    return output_type(indices, new_header)


@overload
def all(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> float:
    ...


@overload
def all(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def all(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def all(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
) -> Union["struct.Map", "struct.Profile", float]:
    """
    Returns True if input.data contains only non-zero elements over the considered axis.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    return _reduce(input, output_type, np.all)


@overload
def any(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> float:
    ...


@overload
def any(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def any(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def any(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
) -> bool:
    """
    Returns True if input.data contains at least one non-zero elements over the considered axis.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    return _reduce(input, output_type, np.any)


@overload
def sum(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> float:
    ...


@overload
def sum(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def sum(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def sum(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", float]:
    """
    Computes a sum over the needed axes to obtain a data of type `output_type`.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if ignore_nans:
        return _reduce(input, output_type, np.nansum)
    return _reduce(input, output_type, np.sum)


@overload
def min(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> float:
    ...


@overload
def min(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def min(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def min(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", float]:
    """
    Computes the minimum value over the needed axes to obtain a data of type `output_type`.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if ignore_nans:
        return _reduce(input, output_type, np.nanmin)
    return _reduce(input, output_type, np.min)


@overload
def max(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> float:
    ...


@overload
def max(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def max(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def max(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", float]:
    """
    Computes the maximum value over the needed axes to obtain a data of type `output_type`.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if ignore_nans:
        return _reduce(input, output_type, np.nanmax)
    return _reduce(input, output_type, np.max)


@overload
def ptp(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> float:
    ...


@overload
def ptp(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def ptp(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def ptp(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", float]:
    """
    Computes the peak-to-peak value over the needed axes to obtain a data of type `output_type`.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    return input.max(output_type, ignore_nans) - input.min(output_type, ignore_nans)


@overload
def argmin(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> tuple:
    ...


@overload
def argmin(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def argmin(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def argmin(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", Tuple[int], int]:
    """
    Computes an argmin over the needed axes to obtain a data of type `output_type`.
    In case of multiple occurrences of the minimum values, the indices corresponding to the first occurrence are returned.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if ignore_nans:
        return _arg(input, output_type, np.nanargmin)
    return _arg(input, output_type, np.argmin)


@overload
def argmax(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> float:
    ...


@overload
def argmax(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def argmax(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def argmax(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", Tuple[int], int]:
    """
    Computes an argmax value over the needed axes to obtain a data of type `output_type`.
    In case of multiple occurrences of the maximum values, the indices corresponding
    to the first occurrence are returned.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if ignore_nans:
        return _arg(input, output_type, np.nanargmax)
    return _arg(input, output_type, np.argmax)


@overload
def mean(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> float:
    ...


@overload
def mean(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def mean(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def mean(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", float]:
    """
    Computes the mean value over the needed axes to obtain a data of type `output_type`.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if ignore_nans:
        return _reduce(input, output_type, np.nanmean)
    return _reduce(input, output_type, np.mean)


@overload
def std(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> float:
    ...


@overload
def std(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def std(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def std(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", float]:
    """
    Computes the standard deviation over the needed axes to obtain a data of type `output_type`.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if ignore_nans:
        return _reduce(input, output_type, np.nanstd)
    return _reduce(input, output_type, np.std)


@overload
def var(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> float:
    ...


@overload
def var(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def var(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def var(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", float]:
    """
    Computes the variance over the needed axes to obtain a data of type `output_type`.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if ignore_nans:
        return _reduce(input, output_type, np.nanvar)
    return _reduce(input, output_type, np.var)


@overload
def moment(
    input: "struct.Struct",
    order: int,
    centered: bool = True,
    output_type: Type[float] = float,
    ignore_nans: bool = True,
) -> float:
    ...


@overload
def moment(
    input: "struct.Struct",
    order: int,
    centered: bool = True,
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def moment(
    input: "struct.Struct",
    order: int,
    centered: bool = True,
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def moment(
    input: "struct.Struct",
    order: int,
    centered: bool = True,
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", float]:
    """
    Computes the statistical moment of order `order` over the needed axes to obtain a data of type `output_type`. If `centered` is True, the centered moment is computed.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    order: int
        Order of the statistical moment.
    centered: bool
        Whether the centered moment is computed. Default: True.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if not isinstance(order, int) or order <= 0:
        raise ValueError(f"order must be a positive integer (here {order})")
    if ignore_nans:
        mean = np.nanmean
    else:
        mean = np.mean
    if centered and isinstance(input, struct.Cube) and output_type is struct.Profile:
        op = lambda a, axis=None: mean(
            (a - np.expand_dims(mean(a, axis=axis), (1, 2))) ** order, axis=axis
        )
    elif centered:
        op = lambda a, axis=None: mean((a - mean(a, axis=axis)) ** order, axis=axis)
    else:
        op = lambda a, axis=None: np.mean(a**order, axis)
    return _reduce(input, output_type, op)


@overload
def median(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> float:
    ...


@overload
def median(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def median(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def median(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", float]:
    """
    Computes median value over the needed axes to obtain a data of type `output_type`.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if ignore_nans:
        return _reduce(input, output_type, np.nanmedian)
    return _reduce(input, output_type, np.median)


@overload
def quantile(
    input: "struct.Struct",
    q: float,
    output_type: Type[float] = float,
    ignore_nans: bool = True,
) -> float:
    ...


@overload
def quantile(
    input: "struct.Struct",
    q: float,
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def quantile(
    input: "struct.Struct",
    q: float,
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def quantile(
    input: "struct.Struct",
    q: float,
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", float]:
    """
    Computes the quantile `q` over the needed axes to obtain a data of type `output_type`.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    q: float
        Quantile (between 0 and 1).
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if ignore_nans:
        op = lambda a, axis=None: np.nanquantile(a, q, axis=axis)
    else:
        op = lambda a, axis=None: np.quantile(a, q, axis=axis)
    return _reduce(input, output_type, op)


@overload
def percentile(
    input: "struct.Struct",
    p: float,
    output_type: Type[float] = float,
    ignore_nans: bool = True,
) -> float:
    ...


@overload
def percentile(
    input: "struct.Struct",
    p: float,
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def percentile(
    input: "struct.Struct",
    p: float,
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def percentile(
    input: "struct.Struct",
    p: float,
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", float]:
    """
    Computes the percentile `p` over the needed axes to obtain a data of type `output_type`.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    p: float
        Percentile (between 0 and 100).
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    if ignore_nans:
        op = lambda a, axis=None: np.nanpercentile(a, p, axis=axis)
    else:
        op = lambda a, axis=None: np.percentile(a, p, axis=axis)
    return _reduce(input, output_type, op)


@overload
def rms(
    input: "struct.Struct", output_type: Type[float] = float, ignore_nans: bool = True
) -> float:
    ...


@overload
def rms(
    input: "struct.Struct",
    output_type: Type["struct.Map"] = float,
    ignore_nans: bool = True,
) -> "struct.Map":
    ...


@overload
def rms(
    input: "struct.Struct",
    output_type: Type["struct.Profile"] = float,
    ignore_nans: bool = True,
) -> "struct.Profile":
    ...


def rms(
    input: "struct.Struct",
    output_type: Type[Union["struct.Map", "struct.Profile", float]] = float,
    ignore_nans: bool = True,
) -> Union["struct.Struct", float]:
    """
    Computes the root mean squared (RMS) value over the needed axes to obtain a data of type `output_type`.

    Parameters
    ----------
    input: struct.Struct
        Input multidimensional data.
    output_type: Type[Map] | Type[Profile] | Type[float]
        Type of output data. Determine the axis over which the operator has to be applied.
    operator:
        Operation on numpy array that reduce the number of dimensions.

    Returns
    -------
    Map | Profile | float
        Resulting data of type `output_type`.
    """
    op = np.nanmean if ignore_nans else np.mean
    return _reduce(
        input, output_type, lambda a, axis=None: np.sqrt(op(a**2, axis=axis))
    )



from typing import overload

import numpy as np
from scipy import ndimage

from .. import structures as struct
from . import filters

__all__ = [
    "StructuringElements",
    "closing",
    "dilation",
    "erosion",
    "filters",
    "gradient",
    "laplacian",
    "opening",
]


class StructuringElements:
    """Helpers to instantiate structuring elements"""

    @staticmethod
    def segment(lg: int, filled: bool = True) -> np.ndarray:
        """TODO"""
        se = np.ones(lg, dtype=np.bool)
        if not filled:
            se[1:-1] = False
        return se

    @staticmethod
    def disk(r: int, filled: bool = True) -> np.ndarray:
        """TODO"""
        x = np.arange(2 * r + 1) - r
        X, Y = np.meshgrid(x, x)
        if filled:
            return (X**2 + Y**2) <= r**2
        return (X**2 + Y**2) == r**2

    @staticmethod
    def ellipse(a: int, b: int, theta: float, filled: bool = True) -> np.ndarray:
        """TODO"""
        raise NotImplementedError()

    @staticmethod
    def square(c: int, filled: bool = True) -> np.ndarray:
        """TODO"""
        return StructuringElements.rectangle(c, c)

    @staticmethod
    def rectangle(a: int, b: int, filled: bool = True) -> np.ndarray:
        """TODO"""
        se = np.ones((a, b), dtype=np.bool)
        if filled:
            se[1:-1, 1:-1] = False
        return se

    @staticmethod
    def ball(r: int, filled: bool = True) -> np.ndarray:
        """TODO"""
        x = np.arange(2 * r + 1) - r
        X, Y, Z = np.meshgrid(x, x, x)
        if filled:
            return (X**2 + Y**2 + Z**2) <= r**2
        return (X**2 + Y**2 + Z**2) == r**2

    @staticmethod
    def ellipsoid(
        rx: int,
        ry: int,
        rz: int,
        theta: float = 0.0,
        phi: float = 0.0,
        filled: bool = True,
    ) -> np.ndarray:  # TODO
        """TODO"""
        r = max(rx, ry, rz)
        x = np.arange(2 * r + 1) - r
        Z, Y, X = np.meshgrid(x, x, x, indexing="ij")
        if filled:
            return (X**2 / rx**2 + Y**2 / ry**2 + Z**2 / rz**2) <= 1
        return (X**2 / rx**2 + Y**2 / ry**2 + Z**2 / rz**2) == 1

    @staticmethod
    def cube(c: int, filled=True) -> np.ndarray:
        """TODO"""
        return StructuringElements.cuboid(c, c, c, filled=filled)

    @staticmethod
    def cuboid(a: int, b: int, c: int, filled: bool = True) -> np.ndarray:
        """TODO"""
        se = np.ones(a, b, c, dtype=np.bool)
        if not filled:
            se[1:-1, 1:-1, 1:-1] = False
        return se


@overload
def dilation(input: "struct.Cube", se: np.ndarray) -> "struct.Cube":
    ...


@overload
def dilation(input: "struct.Map", se: np.ndarray) -> "struct.Map":
    ...


@overload
def dilation(input: "struct.Profile", se: np.ndarray) -> "struct.Profile":
    ...


def dilation(input: "struct.Struct", se: np.ndarray) -> "struct.Struct":
    """TODO"""
    if isinstance(input, struct.Cube):
        if se.ndim == 1:
            return input.filter_pixels(se, filtering_mode="max", padding_mode="reflect")
        if se.ndim == 2:
            return input.filter_channels(
                se, filtering_mode="max", padding_mode="reflect"
            )
        if se.ndim == 3:
            return input.filter(se, filtering_mode="max", padding_mode="reflect")
        raise ValueError(
            f"Structuring element se must have 1, 2 or 3 dimensions, not {se.ndim}, for input of type Cube"
        )
    elif isinstance(input, struct.Map):
        if se.ndim == 2:
            return input.filter(se, filtering_mode="max", padding_mode="reflect")
        raise ValueError(
            f"Structuring element se must have 2 dimensions, not {se.ndim}, for input of type Map"
        )
    elif isinstance(input, struct.Profile):
        if se.ndim == 1:
            return input.filter(se, filtering_mode="max", padding_mode="reflect")
        raise ValueError(
            f"Structuring element se must have 1 dimension, not {se.ndim}, for input of type Profile"
        )
    else:
        raise ValueError(
            f"input must an instance of Cube, Map or Profile, not {type(input)}"
        )


@overload
def erosion(input: "struct.Cube", se: np.ndarray) -> "struct.Cube":
    ...


@overload
def erosion(input: "struct.Map", se: np.ndarray) -> "struct.Map":
    ...


@overload
def erosion(input: "struct.Profile", se: np.ndarray) -> "struct.Profile":
    ...


def erosion(input: "struct.Struct", se: np.ndarray) -> "struct.Struct":
    """TODO"""
    if isinstance(input, struct.Cube):
        if se.ndim == 1:
            return input.filter_pixels(se, filtering_mode="min", padding_mode="reflect")
        if se.ndim == 2:
            return input.filter_channels(
                se, filtering_mode="min", padding_mode="reflect"
            )
        if se.ndim == 3:
            return input.filter(se, "min")
        raise ValueError(
            f"Structuring element se must have 1, 2 or 3 dimensions, not {se.ndim}, for input of type Cube"
        )
    elif isinstance(input, struct.Map):
        if se.ndim == 2:
            return input.filter(se, filtering_mode="min", padding_mode="reflect")
        raise ValueError(
            f"Structuring element se must have 2 dimensions, not {se.ndim}, for input of type Map"
        )
    elif isinstance(input, struct.Profile):
        if se.ndim == 1:
            return input.filter(se, filtering_mode="min", padding_mode="reflect")
        raise ValueError(
            f"Structuring element se must have 1 dimension, not {se.ndim}, for input of type Profile"
        )
    else:
        raise ValueError(
            f"input must an instance of Cube, Map or Profile, not {type(input)}"
        )


@overload
def closing(input: "struct.Cube", se: np.ndarray) -> "struct.Cube":
    ...


@overload
def closing(input: "struct.Map", se: np.ndarray) -> "struct.Map":
    ...


@overload
def closing(input: "struct.Profile", se: np.ndarray) -> "struct.Profile":
    ...


def closing(input: "struct.Struct", se: np.ndarray) -> "struct.Struct":
    """TODO"""
    return erosion(dilation(input, se), se)


@overload
def opening(input: "struct.Cube", se: np.ndarray) -> "struct.Cube":
    ...


@overload
def opening(input: "struct.Map", se: np.ndarray) -> "struct.Map":
    ...


@overload
def opening(input: "struct.Profile", se: np.ndarray) -> "struct.Profile":
    ...


def opening(input: "struct.Struct", se: np.ndarray) -> "struct.Struct":
    """TODO"""
    return dilation(erosion(input, se), se)


@overload
def gradient(input: "struct.Cube", se: np.ndarray) -> "struct.Cube":
    ...


@overload
def gradient(input: "struct.Map", se: np.ndarray) -> "struct.Map":
    ...


@overload
def gradient(input: "struct.Profile", se: np.ndarray) -> "struct.Profile":
    ...


def gradient(input: "struct.Struct", se: np.ndarray) -> "struct.Struct":
    """TODO"""
    return dilation(input, se) - erosion(input, se)


@overload
def laplacian(input: "struct.Cube", se: np.ndarray) -> "struct.Cube":
    ...


@overload
def laplacian(input: "struct.Map", se: np.ndarray) -> "struct.Map":
    ...


@overload
def laplacian(input: "struct.Profile", se: np.ndarray) -> "struct.Profile":
    ...


def laplacian(input: "struct.Struct", se: np.ndarray) -> "struct.Struct":
    """TODO"""
    return dilation(input, se) + erosion(input, se) - 2 * input



from math import sqrt
from numbers import Number
from typing import Literal, Union, overload

import numpy as np

from .. import structures as struct

__all__ = ["additive_noise", "multiplicative_noise"]


def _std_broadcasting(
    input: "struct.Struct", std: Union["struct.Struct", float] = 1.0
) -> np.ndarray:
    std = std if std is not None else 1
    if isinstance(input, struct.Cube):
        if isinstance(std, struct.Cube):
            std_ = std.data
        elif isinstance(std, struct.Map):
            std_ = std.data[np.newaxis, :, :] * np.ones((input.nz, 1, 1))
        elif isinstance(std, struct.Profile):
            std_ = std.data[:, np.newaxis, np.newaxis] * np.ones(
                (1, input.ny, input.nx)
            )
        elif isinstance(std, Number):
            std_ = std * np.ones((input.nz, input.ny, input.nx))
        else:
            raise ValueError(
                f"std is of type {type(std)} which is incompatible for input of type {type(input)}"
            )
    elif isinstance(input, struct.Map):
        if isinstance(std, struct.Map):
            std_ = std.data
        elif isinstance(std, Number):
            std_ = std * np.ones((input.ny, input.nx))
        else:
            raise ValueError(
                f"std is of type {type(std)} which is incompatible for input of type {type(input)}"
            )
    elif isinstance(input, struct.Profile):
        if isinstance(std, struct.Profile):
            std_ = std.data
        elif isinstance(std, Number):
            std_ = std * np.ones(input.nz)
        else:
            raise ValueError(
                f"std is of type {type(std)} which is incompatible for input of type {type(input)}"
            )
    else:
        raise ValueError(
            f"input must be of type Cube, Map or Profile, not {type(input)}"
        )
    return std_


@overload
def additive_noise(
    input: "struct.Cube",
    noise_type: Literal["gaussian", "uniform"],
    std: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def additive_noise(
    input: "struct.Map",
    noise_type: Literal["gaussian", "uniform"],
    std: Union["struct.Map", float],
) -> "struct.Map":
    ...


@overload
def additive_noise(
    input: "struct.Profile",
    noise_type: Literal["gaussian", "uniform"],
    std: Union["struct.Profile", float],
) -> "struct.Profile":
    ...


def additive_noise(
    input: "struct.Struct",
    noise_type: Literal["gaussian", "uniform"],
    std: Union["struct.Struct", float] = 1.0,
) -> "struct.Struct":
    """
    Return the input structure degraded with an additive noise of type `noise_type`.

    Parameters
    ----------
    input : Cube | Map | Profile
        Input structure.
    noise_type : str
        Type of noise. Must be 'gaussian' or 'uniform'.
    std : Cube | Map | Profile
        Standard deviation of noise.

    Returns
    -------

    res : Cube | Map | Profile
        Noisy structure.
    """
    noise_type = noise_type.lower().strip()
    std_ = _std_broadcasting(input, std)

    if noise_type == "gaussian":
        data = input.data + np.random.normal(0, std_)
    elif noise_type == "uniform":
        data = input.data + np.random.uniform(-sqrt(3) * std_, sqrt(3) * std_)
    else:
        raise ValueError(f"noise_type must be gaussian or uniform, not {noise_type}")

    return type(input)(data, input.header)


@overload
def multiplicative_noise(
    input: "struct.Cube",
    noise_type: Literal["gaussian", "uniform"],
    std: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def multiplicative_noise(
    input: "struct.Map",
    noise_type: Literal["gaussian", "uniform"],
    std: Union["struct.Map", float],
) -> "struct.Map":
    ...


@overload
def multiplicative_noise(
    input: "struct.Profile",
    noise_type: Literal["gaussian", "uniform"],
    std: Union["struct.Profile", float],
) -> "struct.Profile":
    ...


def multiplicative_noise(
    input: "struct.Struct",
    noise_type: Literal["gaussian", "uniform"],
    std: Union["struct.Struct", float] = 1.0,
) -> "struct.Struct":
    """
    Return the input structure degraded with a multiplicative noise of type `noise_type`.

    Parameters
    ----------
    input : Cube | Map | Profile
        Input structure.
    noise_type : str
        Type of noise. Must be 'gaussian' or 'uniform'.
    std : Cube | Map | Profile
        Standard deviation of noise.

    Returns
    -------

    res : Cube | Map | Profile
        Noisy structure.
    """
    noise_type = noise_type.lower().strip()
    std_ = _std_broadcasting(input, std)

    if noise_type == "gaussian":
        data = input.data * (1 + np.random.normal(0, std_))
    elif noise_type == "uniform":
        data = input.data * (1 + np.random.uniform(-sqrt(3) * std_, sqrt(3) * std_))
    else:
        raise ValueError(f"noise_type must be gaussian or uniform, not {noise_type}")

    return type(input)(data, input.header)



""" Header module """

from typing import Optional, Union, Literal

from datetime import datetime

import numpy as np

from astropy.io import fits

__all__ = [
    'C_LIGHT', 'bound_coordinates', 'bound_frequencies', 'bound_velocities', 'change_coordinates', 'check_header', 'coordinates_to_indices', 'create_header', 'frequency_to_index', 'index_to_frequency', 'index_to_velocity', 'indices_to_coordinates', 'merge_headers', 'move_header_axes', 'remove_header_axis', 'update_header', 'velocity_to_index'
]

def check_header(data, header) :
    """ Returns True if data is compatible with the header, else raise a `ValueError` """
    if header['NAXIS'] != data.ndim :
        raise ValueError(f"header must have card NAXIS = {data.ndim}, not {header['NAXIS']}")
    h_shape = tuple((header[f'NAXIS{k}'] for k in range(data.ndim, 0, -1)))
    if data.shape != h_shape :
        raise ValueError(f"data dimensions {data.shape} do not match the header cards {h_shape}")
    return True

def update_header(data : np.ndarray, header : fits.Header) :
    """ Updates header cards to match data. Data and header are first verified by `check_header`
    i.e. their dimensions must match. Missing cards are also added to avoid future errors. """
    check_header(data, header)
    new_header = header.copy()
    # Update data extrema (if all values are NaNs, min and max are 0 by default)
    if np.isfinite(data).any() :
        new_header['DATAMIN'] = np.nanmin(data)
        new_header['DATAMAX'] = np.nanmax(data)
    else :
        new_header['DATAMIN'] = 0.
        new_header['DATAMAX'] = 0.
    # Update date
    new_header['DATE'] = str(datetime.now())
    # Update origin
    if 'ORIGIN' in new_header.keys() :
        if '- MODIFIED' not in new_header['ORIGIN'].strip().upper() :
            new_header['ORIGIN'] = new_header['ORIGIN'].strip().upper() + ' - MODIFIED'
    # Update several cards
    if 'OBJECT' not in new_header.keys() :
        new_header['OBJECT'] = 'UNKNOWN'
    if 'BUNIT' not in new_header.keys() :
        new_header['BUNIT'] = 'UNKNOWN'
    if 'RA' not in new_header.keys() :
        new_header['RA'] = 0.
    if 'DEC' not in new_header.keys() :
        new_header['DEC'] = 0.
    if 'RESTFREQ' not in new_header.keys() :
        new_header['RESTFREQ'] = 1.
    # Axis cards
    if data.ndim not in (1, 2, 3) :
        raise ValueError(f'data dimension must be 1, 2 or 3, not {data.ndim}')
    for axis in range(1, data.ndim+1) :
        if f'CTYPE{axis}' not in new_header.keys() :
            new_header[f'CTYPE{axis}'] = 'INDICES'
        if f'CRVAL{axis}' not in new_header.keys() :
            new_header[f'CRVAL{axis}'] = 1
        if f'CDELT{axis}' not in new_header.keys() :
            new_header[f'CDELT{axis}'] = 1
        if f'CRPIX{axis}' not in new_header.keys() :
            new_header[f'CRPIX{axis}'] = 1
        if f'CROTA{axis}' not in new_header.keys() :
            new_header[f'CROTA{axis}'] = 0
    return new_header

def merge_headers(header_1 : fits.Header, header_2 : fits.Header) :
    """
    Returns a header compatible with `header_1` and `header_2`.
    The returned header can have a different number of axis than the inputs.
    """
    n1, n2 = header_1['NAXIS'], header_2['NAXIS']
    cards = ['NAXIS{}', 'CTYPE{}', 'CRVAL{}', 'CDELT{}', 'CRPIX{}', 'CROTA{}']

    if n1 == n2 :
        # Case 1 : header have the same number of axis
        h_shape_1 = tuple((header_1[f'NAXIS{k}'] for k in range(header_1['NAXIS'], 0, -1)))
        h_shape_2 = tuple((header_2[f'NAXIS{k}'] for k in range(header_1['NAXIS'], 0, -1)))
        if h_shape_1 != h_shape_2 :
            raise ValueError(f"header_1 and header_2 must have the same axis dimensions ({h_shape_1} and {h_shape_2})")
        # By default, the new header is a copy of header_1
        new_header = header_1.copy()
    elif (n1 == 2 and n2 == 1) or (n1 == 1 and n2 == 2) :
        # Case 2 : one has two spatial axis and the other has one spectral axis
        (map_header, profile_header) = (header_1.copy(), header_2.copy())\
            if n1 > n2 else (header_2.copy(), header_1.copy())
        map_header['NAXIS'] = 3
        for card in cards :
            map_header[card.format(3)] = profile_header[card.format(1)]
        new_header = map_header
    elif (n1 == 3 and n2 == 2) or (n1 == 2 and n2 == 3) :
        # Case 3 : one has three axis and the other has two spatial axis
        # By default, the new header is a copy of the cube header
        new_header = header_1.copy() if n1 > n2 else header_2.copy()
    elif (n1 == 3 and n2 == 1) or (n1 == 1 and n2 == 3) :
        # Case 4 : one has three axis and the other has one spectral axis
        # By default, the new header is a copy of the profile header
        new_header = header_1.copy() if n1 > n2 else header_2.copy()
    else :
        # Else return an error
        raise ValueError("Headers cannot be merged.")

    # Check somes cards
    try :
        if header_1['LINE'].strip().lower() != header_2['LINE'].strip().lower() :
            new_header['LINE'] = 'COMBINATION'
    except :
        pass
    # Remove unit
    if 'BUNIT' in new_header.keys() :
        new_header['BUNIT'] = 'UNKNOWN'
    # Update origin
    if 'ORIGIN' in new_header.keys() :
        if '- MODIFIED' not in new_header['ORIGIN'].strip().upper() :
            new_header['ORIGIN'] = new_header['ORIGIN'].strip().upper() + ' - MODIFIED'

    return new_header

def create_header(struct : str, ref_header : Optional[fits.Header] = None,
    nx : Optional[int] = None, ny : Optional[int] = None, nz : Optional[int] = None) -> fits.Header:
    """
    TODO
    """
    struct = struct.lower()
    if struct not in ['cube', 'map', 'profile'] :
        raise ValueError(f"struct must be 'cube', 'map' or 'profile', not {struct}")

    if ref_header is None :
        new_header = fits.Header()
        new_header['SIMPLE'] = 'T'
        new_header['BITPIX'] = 32     
        #new_header['BSCALE'] = 0.3341172671867E-08
        #new_header['BZERO'] = 0.3305371761322E+01
        #new_header['BLANK'] = 2147483647
        new_header['DATAMIN'] = 0.
        new_header['DATAMAX'] = 0.
        new_header['BUNIT'] = 'UNKNOWN'
        new_header['OBJECT'] = 'UNKNOWN'
        new_header['RADESYS'] = 'FK5'
        #new_header['RA'] =  0.8522612499999E+02
        #new_header['DEC'] = -0.2466666666667E+01
        #new_header['EQUINOX'] =  0.2000000000000E+04
        new_header['LINE'] = 'UNKNOWN'
        #new_header['ALTRPIX'] = 0.1200000000000E+03
        #new_header['ALTRVAL'] = 0.1134869952662E+12
        #new_header['RESTFREQ'] = 0.1134909702000E+12
        #new_header['IMAGFREQ'] = 0.9400874075270E+11
        #new_header['VELO-LSR'] = 0.1050000000000E+05
        #new_header['VELREF'] = 257
        #new_header['SPECSYS'] = 'LSRK'
        #new_header['BMAJ'] =  0.6346730671426E-02
        #new_header['BMIN'] =  0.6346730671426E-02
        #new_header['BPA'] =  0.0000000000000E+00
        new_header['ORIGIN'] = 'IRAM'
        new_header['DATE'] = str(datetime.now())
        if struct == 'cube' :
            if nx is None or ny is None or nz is None :
                raise ValueError('nx, ny and nz cannot be None to create a cube header without reference')
            new_header['NAXIS'] = 3
            new_header['NAXIS1'] = nx
            new_header['NAXIS2'] = ny
            new_header['NAXIS3'] = nz
            new_header['CTYPE1'] = 'INDICES'
            new_header['CTYPE2'] = 'INDICES'
            new_header['CTYPE3'] = 'INDICES'
            new_header['CRVAL1'] = 1
            new_header['CDELT1'] = 1
            new_header['CRPIX1'] = 1
            new_header['CROTA1'] = 0
            new_header['CRVAL2'] = 1
            new_header['CDELT2'] = 1
            new_header['CRPIX2'] = 1
            new_header['CROTA2'] = 0
            new_header['CRVAL3'] = 1
            new_header['CDELT3'] = 1
            new_header['CRPIX3'] = 1
            new_header['CROTA3'] = 0
            return new_header
        if struct == 'map' :
            if nx is None or ny is None :
                raise ValueError('nx and ny cannot be None to create a map header without reference')
            new_header['NAXIS'] = 2
            new_header['NAXIS1'] = nx
            new_header['NAXIS2'] = ny
            new_header['CTYPE1'] = 'INDICES'
            new_header['CTYPE2'] = 'INDICES'
            new_header['CRVAL1'] = 1
            new_header['CDELT1'] = 1
            new_header['CRPIX1'] = 1
            new_header['CROTA1'] = 0
            new_header['CRVAL2'] = 1
            new_header['CDELT2'] = 1
            new_header['CRPIX2'] = 1
            new_header['CROTA2'] = 0
            return new_header
        if struct == 'profile' :
            if nz is None :
                raise ValueError('nz cannot be None to create a profile header without reference')
            new_header['NAXIS'] = 1
            new_header['NAXIS1'] = nz
            new_header['CTYPE1'] = 'INDICES'
            new_header['CRVAL1'] = 1
            new_header['CDELT1'] = 1
            new_header['CRPIX1'] = 1
            new_header['CROTA1'] = 0
            return new_header

    if ref_header['NAXIS'] == 1 :
        ref_struct = 'profile'
    elif ref_header['NAXIS'] == 2 :
        ref_struct = 'map'
    elif ref_header['NAXIS'] == 3 :
        ref_struct = 'cube'
    else :
        raise ValueError(f"ref_header['NAXIS'] must be equal to 1, 2 or 3, not {ref_header['NAXIS']}")

    if struct == 'cube' and ref_struct == 'cube' :
        return ref_header.copy()
    if struct == 'cube' and ref_struct == 'map' :
        return merge_headers(ref_header, create_header('profile', nz = nz))
    if struct == 'cube' and ref_struct == 'profile' :
        return merge_headers(ref_header, create_header('map', nx = nx, ny = ny))
    
    raise ValueError(f'Cannot create {struct} header using a {ref_struct} header as reference')
    

# Cards

def move_header_axes(header : fits.Header, order : Literal['xyz', 'yxz', 'zxy', 'zyx', 'xy', 'yx']) -> fits.Header :
    """
    Return a version of header where the order of axes if xyz.
    Input header axes order is `order`.
    """
    order = order.lower()

    # If header is a cube header
    if header['NAXIS'] == 3 :
        if order == 'xyz' :
            indices = (1, 2, 3)
        elif order == 'yxz' :
            indices = (2, 1, 3)
        elif order == 'zxy' :
            indices = (3, 1, 2)
        elif order == 'zyx' :
            indices = (3, 2, 1)
        else :
            raise ValueError(f"order must be 'xyz', 'yxz', 'zxy' or 'zyx', not '{order}'")
    elif header['NAXIS'] == 2 :
        if order == 'xy' :
            indices = (1, 2)
        elif order == 'yx' :
            indices = (2, 1)
        else :
            raise ValueError(f"order must be 'xy' or 'yx', not '{order}'")
    else :
        raise ValueError('move_header_axes is not defined for a number of axis different than 2 or 3')
    
    new_header = header.copy()
    cards = ['NAXIS{}', 'CTYPE{}', 'CRVAL{}', 'CDELT{}', 'CRPIX{}', 'CROTA{}']
    for i, j in enumerate(indices, 1) :
        for card in cards :
            new_header[card.format(j)] = header[card.format(i)]
    return new_header

def remove_header_axis(header : fits.Header, axis : str) -> fits.Header :
    """
    Axis must be 'spatial' or 'spectral'.
    Header must be a cube header i.e. have header['NAXIS'] = 3.
    """
    axis = axis.strip().lower()
    cards = ['NAXIS{}', 'CTYPE{}', 'CRVAL{}', 'CDELT{}', 'CRPIX{}', 'CROTA{}']
    new_header = header.copy()
    if axis == 'spatial' :
        for card in cards :
            new_header[card.format(1)] = new_header[card.format(3)]
            del new_header[card.format(2)]
            del new_header[card.format(3)]
        new_header['NAXIS'] = 1
    elif axis == 'spectral' :
        for card in cards :
            del new_header[card.format(3)]
        new_header['NAXIS'] = 2
    else :
        raise ValueError(f"axis must be 'spatial' or 'spectral', not {axis}")
    return new_header

def _replace_none(x, value) :
    """ Returns x+1 if x is not None, else value """
    return x+1 if x is not None else value

def change_coordinates(header, x = None, y = None, z = None) -> fits.Header :
    """ Coordinates must begin at 0 """
    new_header = header.copy()

    if header['NAXIS'] == 3 : # Cubes handling

        if x is None :
            x = [1, header['NAXIS1']]
        else :
            x = [_replace_none(x[0], 1), _replace_none(x[1], header['NAXIS1'])]
        if y is None :
            y = [1, header['NAXIS2']]
        else :
            y = [_replace_none(y[0], 1), _replace_none(y[1], header['NAXIS2'])]
        if z is None :
            z = [1, header['NAXIS3']]
        else :
            z = [_replace_none(z[0], 1), _replace_none(z[1], header['NAXIS3'])]

        new_header['CRPIX1'] = new_header['CRPIX1'] - x[0]
        new_header['CRPIX2'] = new_header['CRPIX2'] - y[0]
        new_header['CRPIX3'] = new_header['CRPIX3'] - z[0]

        new_header['NAXIS1'] = x[1] - x[0] + 1
        new_header['NAXIS2'] = y[1] - y[0] + 1
        new_header['NAXIS3'] = z[1] - z[0] + 1

    elif header['NAXIS'] == 2 : # Maps handling

        if x is None :
            x = [1, header['NAXIS1']]
        else :
            x = [_replace_none(x[0], 1), _replace_none(x[1], header['NAXIS1'])]
        if y is None :
            y = [1, header['NAXIS2']]
        else :
            y = [_replace_none(y[0], 1), _replace_none(y[1], header['NAXIS2'])]

        new_header['CRPIX1'] = new_header['CRPIX1'] - x[0]
        new_header['CRPIX2'] = new_header['CRPIX2'] - y[0]

        new_header['NAXIS1'] = x[1] - x[0] + 1
        new_header['NAXIS2'] = y[1] - y[0] + 1

    elif header['NAXIS'] == 1 : # Profiles handling

        if z is None :
            z = [1, header['NAXIS1']]
        else :
            z = [_replace_none(z[0], 1), _replace_none(z[1], header['NAXIS1'])]

        new_header['CRPIX1'] = new_header['CRPIX1'] - z[0]

        new_header['NAXIS1'] = z[1] - z[0] + 1

    else :

        raise ValueError(f"'NAXIS' card must be 1, 2 or 3 (not {header['NAXIS']})")
   
    return new_header

# Coordinates
# Unit for index is numpy index (beginning at 0)
# Unit for velocity is km/s
# Unit for frequency is GHz
# Unit for angle is ' (minute of arc)

C_LIGHT = 299792458 # In m/s

def indices_to_coordinates(header : fits.Header, i : Union[int, np.ndarray],
    j : Union[int, np.ndarray], absolute : bool = False) :
    """ Returns the absolute coordinates in degrees of the (i,j) point in pixels (beginning at 0) """
    if header['NAXIS'] not in [2, 3] : # If header of profile
        raise ValueError("header must be the header of a cube or a map")
    if (np.array(i) < 0).any() or (np.array(j) < 0).any() :
        raise ValueError('i and j must be non-negative indices')
    if (np.array(i) >= header['NAXIS1']).any() :
        raise ValueError(f"i must be lower than {header['NAXIS1']}")
    if (np.array(j) >= header['NAXIS2']).any() :
        raise ValueError(f"j must be lower than {header['NAXIS2']}")

    if absolute :
        x = (i+1-header['CRPIX1']) * header['CDELT1'] + header['CRVAL1']
        y = (j+1-header['CRPIX2']) * header['CDELT2'] + header['CRVAL2']
    else :
        x = (i+1-header['CRPIX1']) * header['CDELT1']
        y = (j+1-header['CRPIX2']) * header['CDELT2']
    x, y = 60*x, 60*y # Conversion from degrees to minutes of arc
    return x, y

def coordinates_to_indices(header : fits.Header, x : Union[float, np.ndarray],
    y : Union[float, np.ndarray], absolute : bool = False) :
    """ TODO """
    if header['NAXIS'] == 1 :
        raise ValueError("header must be the header of a cube or a map")
    xmin, ymin = indices_to_coordinates(header, 0, 0, absolute = absolute)
    xmax, ymax = indices_to_coordinates(header, header['NAXIS1']-1, header['NAXIS2']-1, absolute = absolute)
    if (np.array(x) < xmin).any() :
        raise ValueError(f"x must be greater than {xmin}'")
    if (np.array(y) < ymin).any() :
        raise ValueError(f"y must be greater than {ymin}'")
    if (np.array(x) > xmax).any() :
        raise ValueError(f"x must be lower than {xmax}'")
    if (np.array(y) > ymax).any() :
        raise ValueError(f"y must be lower than {ymax}'")

    if absolute :
        i = (x - header['CRVAL1']) / header['CDELT1'] + header['CRPIX1'] - 1
        j = (y - header['CRVAL2']) / header['CDELT2'] + header['CRPIX2'] - 1
    else :
        i = x / header['CDELT1'] + header['CRPIX1'] - 1
        j = y / header['CDELT2'] + header['CRPIX2'] - 1
    return round(i), round(j)

def bound_coordinates(header : fits.Header, absolute : bool = False) :
    """ Returns the absolute coordinates bounds """
    xymin = indices_to_coordinates(header, 0, 0, absolute = absolute)
    xymax = indices_to_coordinates(header, header['NAXIS1']-1, header['NAXIS2']-1, absolute = absolute)
    xbounds = xymin[0], xymax[0]
    ybounds = xymin[1], xymax[1]
    return xbounds, ybounds


def index_to_velocity(header : fits.Header, k : Union[int, np.ndarray]) :
    """ Returns the velocity of the k-th spectral image by Doppler effect (beginning at 0) """
    if header['NAXIS'] not in [1, 3] : # If header of map
        raise ValueError("header must be the header of a cube or a profile")
    if (np.array(k) < 0).any() :
        raise ValueError('k must be a non-negative index')
    
    if header['NAXIS'] == 1 : # Header of profile
        if (np.array(k) >= header['NAXIS1']).any() :
            raise ValueError(f"k must be lower than {header['NAXIS1']}")
        v = (k+1-header['CRPIX1']) * header['CDELT1'] + header['CRVAL1']
    else : # Header of cube
        if (np.array(k) >= header['NAXIS3']).any() :
            raise ValueError(f"k must be lower than {header['NAXIS3']}")
        v = (k+1-header['CRPIX3']) * header['CDELT3'] + header['CRVAL3']
    return v / 1e3 # Convert to km/s

def velocity_to_index(header : fits.Header, v : Union[float, np.ndarray]) :
    """ Returns the velocity channel index (beginning at 0) corresponding to a velocity v  """
    if header['NAXIS'] not in [1, 3] :
        raise ValueError("header must be the header of a cube or a profile")
    vmin = index_to_velocity(header, 0)
    if (np.array(v) < vmin).any() :
        raise ValueError(f'v must be greater than {vmin}')

    if header['NAXIS'] == 1 : # Header of profile
        vmax = index_to_velocity(header, header['NAXIS1']-1)
        if (np.array(v) > vmax).any() :
            raise ValueError(f'v must be lower than {vmax}')
        v_ = v * 1e3 # Convert to m/s
        k = (v_ - header['CRVAL1']) / header['CDELT1'] + header['CRPIX1'] - 1
    else : # Header of cube
        vmax = index_to_velocity(header, header['NAXIS3']-1)
        if (np.array(v) > vmax).any() :
            raise ValueError(f'v must be lower than {vmax}')
        v_ = v * 1e3 # Convert to m/s
        k = (v_ - header['CRVAL3']) / header['CDELT3'] + header['CRPIX3'] - 1
    return round(k)

def bound_velocities(header : fits.Header) :
    """ Returns the min and max velocities """
    vmin = index_to_velocity(header, 0)
    vmax = index_to_velocity(header, header[f"NAXIS{header['NAXIS']}"]-1)
    return vmin, vmax
    

def index_to_frequency(header : fits.Header, k : Union[int, np.ndarray]) :
    """ Returns the frequency of the k-th spectal image (beginning at 0) """
    if header['NAXIS'] not in [1, 3] : # If header of map
        raise ValueError("header must be the header of a cube or a profile")
    if (np.array(k) < 0).any() :
        raise ValueError('k must be a non-negative index')
    
    if header['NAXIS'] == 1 :
        if (np.array(k) >= header['NAXIS1']).any() :
            raise ValueError(f"k must be lower than {header['NAXIS1']}")
        df = -(header['CDELT1']/C_LIGHT) * header['RESTFREQ']
        f = (k+1-header['CRPIX1']) * df + header['RESTFREQ']
    else :
        if (np.array(k) >= header['NAXIS3']).any() :
            raise ValueError(f"k must be lower than {header['NAXIS3']}")
        df = -(header['CDELT3']/C_LIGHT) * header['RESTFREQ']
        f = (k+1-header['CRPIX3']) * df + header['RESTFREQ']
    return f / 1e9 # Convert to GHz

def frequency_to_index(header : fits.Header, f : Union[float, np.ndarray]) :
    """ Returns the velocity channel index (beginning at 0) corresponding to a frequency f  """
    if header['NAXIS'] not in [1, 3] :
        raise ValueError("header must be the header of a cube or a profile")
    fmin = index_to_frequency(header, 0)
    if (np.array(f) < fmin).any() :
        raise ValueError(f'f must be greater than {fmin}')

    if header['NAXIS'] == 1 :
        fmax = index_to_velocity(header, header['NAXIS1']-1)
        if (np.array(f) > fmax).any() :
            raise ValueError(f'f must be lower than {fmax}')
        f_ = f * 1e9 # Convert to Hz
        df = -(header['CDELT1']/C_LIGHT) * header['RESTFREQ']
        k = (f_ - header['RESTFREQ']) / df + header['CRPIX1'] - 1
    else :
        fmax = index_to_velocity(header, header['NAXIS3']-1)
        if (np.array(f) > fmax).any() :
            raise ValueError(f'f must be lower than {fmax}')
        f_ = f * 1e9 # Convert to Hz
        df = -(header['CDELT3']/C_LIGHT) * header['RESTFREQ']
        k = (f_ - header['RESTFREQ']) / df + header['CRPIX3'] - 1
    return round(k)

def bound_frequencies(header : fits.Header) :
    """ Returns the min and max frequencies """
    fmin = index_to_frequency(header, header[f"NAXIS{header['NAXIS']}"]-1)
    fmax = index_to_frequency(header, 0)
    return fmin, fmax


from typing import Union, Sequence, Tuple, List

from .. import structures as struct
from ..core import header as hdr

__all__ = [
    'get_channel', 'get_channels', 'get_pixel', 'get_pixels'
]

def get_channel(cube : 'struct.Cube', z : Union[int, float],
    unit = 'index') -> 'struct.Map' :
    """ TODO """
    return cube.get_channels([z], unit = unit)[0]

def get_channels(cube : 'struct.Cube', z : Sequence[Union[int, float]],
    unit = 'index') -> List['struct.Map'] :
    """ TODO """
    unit = unit.lower()
    if unit == 'index' :
        k_ = z
    elif unit == 'velocity' :
        k_ = [hdr.velocity_to_index(cube.header, v) for v in z]
    elif unit == 'frequency' :
        k_ = [hdr.frequency_to_index(cube.header, f) for f in z]
    else :
        raise ValueError(f"unit must be 'index', 'velocity' or 'frequency', not '{unit}'")
    
    new_header = hdr.remove_header_axis(cube.header, 'spectral')
    out = []
    for k in k_ :
        new_data = cube.data[k]
        out.append(struct.Map(new_data, new_header))
    return out

def get_pixel(cube : 'struct.Cube', xy : Tuple[Union[int, float]],
    unit = 'index') -> 'struct.Profile' :
    """ TODO """
    return cube.get_pixels([xy], unit = unit)[0]

def get_pixels(cube : 'struct.Cube', xy : Sequence[Tuple[Union[int, float]]],
    unit = 'index') -> List['struct.Profile'] :
    """ TODO """
    unit = unit.lower()
    if unit == 'index' :
        ij_ = xy
    elif unit == 'angle' :
        ij_ = [hdr.coordinates_to_indices(cube.header, x, y) for (x,y) in xy ]
    else :
        raise ValueError(f"unit must be 'index' or 'angle', not '{unit}'")

    new_header = hdr.remove_header_axis(cube.header, 'spatial')
    out = []
    for (i,j) in ij_ :
        new_data = cube.data[:, j, i]
        out.append(struct.Profile(new_data, new_header))
    return out



from typing import Optional, Sequence, Union

import numpy as np

from .. import structures as struct
from ..core import header as hdr

__all__ = [
    "EXCLUDED",
    "TRAIN",
    "VAL",
    "extract_indices",
    "nan_mask",
    "random_spatial_splitting",
    "random_splitting",
    "spatial_splitting",
    "stack_indices",
]

TRAIN = 0
VAL = 1
EXCLUDED = -1

# NaNs detection


def nan_mask(input: "struct.Struct", item: str) -> "struct.Struct":
    """
    Returns a binary mask equal to 1 if an item contains at least a NaN, else 0.
    """
    if item.strip().lower() not in ["pixel", "channel"]:
        raise ValueError(f"item must be 'pixel' or 'channel', not {item}")
    if isinstance(input, struct.Cube):
        if item == "pixel":
            new_data = np.isnan(input.data).max(axis=0)
            new_header = hdr.remove_header_axis(input.header, axis="spectral")
            return struct.Map(new_data, new_header)
        else:
            new_data = np.isnan(input.data).max(axis=(1, 2))
            new_header = hdr.remove_header_axis(input.header, axis="spatial")
            return struct.Map(new_data, new_header)
    elif isinstance(input, struct.Map):
        if item == "pixel":
            new_data = np.isnan(input.data)
            return struct.Map(new_data, input.header)
        else:
            raise ValueError("item = 'channel' is not compatible with a Map inputect")
    elif isinstance(input, struct.Profile):
        if item == "pixel":
            raise ValueError("item = 'pixel' is not compatible with a Profile inputect")
        else:
            new_data = np.isnan(input.data)
            return struct.Profile(new_data, input.header)
    else:
        raise TypeError(
            f"input must be an instance of Cube, Map or Profile, not {type(input)}"
        )


def _common_nan_mask(
    inputs: Union["struct.Struct", Sequence["struct.Struct"]], item: str
) -> "struct.Struct":
    """
    Returns a binary mask equal to 1 if an item contains at least a NaN, else 0.
    """
    if not isinstance(inputs, Sequence):
        inputs = [inputs]
    nans = nan_mask(inputs[0], item)
    for input in inputs[1:]:
        nans = nans | nan_mask(input, item)
    return nans


# Splitting methods


def random_splitting(
    inputs: Union["struct.Struct", Sequence["struct.Struct"]],
    item: str,
    frac_train: float,
    seed: Optional[int] = None,
    reject_nans: bool = True,
) -> "struct.Struct":
    """
    Returns a ternary structure equal to 1 if an item is in the training set,
    0 if it is in the validation set and -1 if the item contains at least one NaN.
    The splitting structure is valid for every inputect in inputs.
    """
    np.random.seed(seed)
    nans = _common_nan_mask(inputs, item)
    if reject_nans:
        indices = np.arange(nans.data.size)[(~nans.data.astype("bool")).flatten()]
    else:
        indices = np.arange(nans.data.size)
    np.random.shuffle(indices)
    train_indices = indices[: int(frac_train * len(indices))]
    val_indices = indices[int(frac_train * len(indices)) :]
    splitting_map = np.zeros_like(nans.data) + EXCLUDED
    if isinstance(nans, struct.Map):
        cols = splitting_map.shape[1]
        splitting_map[train_indices // cols, train_indices % cols] = TRAIN
        splitting_map[val_indices // cols, val_indices % cols] = VAL
    elif isinstance(nans, struct.Profile):
        splitting_map[train_indices] = TRAIN
        splitting_map[val_indices] = VAL
    else:
        raise RuntimeError("ERROR : should never been here")
    return type(nans)(splitting_map, nans.header)


def spatial_splitting(
    inputs: Union["struct.Struct", Sequence["struct.Struct"]], item: str
):
    """
    Returns a ternary structure equal to 1 if an item is in the training set,
    0 if it is in the validation set and -1 if the item contains at least one NaN
    """
    raise NotImplementedError()


def random_spatial_splitting(
    inputs: Union["struct.Struct", Sequence["struct.Struct"]], item: str
):
    """
    Returns a ternary structure equal to 1 if an item is in the training set,
    0 if it is in the validation set and -1 if the item contains at least one NaN
    """
    raise NotImplementedError()


# Indices extraction


def extract_indices(input: Union["struct.Map", "struct.Profile"]) -> tuple[np.ndarray]:
    """
    Returns the indices of the training, validation and full sets.
    """
    if not isinstance(input, (struct.Map, struct.Profile)):
        raise TypeError(f"input must be of type Map or Profile, not {type(input)}")
    data = input.data.flatten()
    indices = np.arange(data.size)
    train_indices = indices[data == TRAIN]
    val_indices = indices[data == VAL]
    return train_indices, val_indices, indices


def stack_indices(extracted: Sequence[tuple[np.ndarray]]) -> tuple[np.ndarray]:
    """
    Returns the stacked indices of the training, validation and full sets.
    """
    indices = np.array([])
    train_indices = np.array([])
    val_indices = np.array([])
    offset = 0
    for (train_ind, val_ind, ind) in extracted:
        offset += ind.max()
        indices = np.concatenate(indices, ind + offset)
        train_indices = np.concatenate(train_indices, train_ind + offset)
        val_indices = np.concatenate(val_indices, val_ind + offset)
    return train_indices, val_indices, indices



from typing import Optional, overload

import numpy as np

from .. import structures as struct

__all__ = [
    "abs",
    "arccos",
    "arccosh",
    "arcsin",
    "arcsinh",
    "arctan",
    "arctanh",
    "cbrt",
    "cos",
    "cosh",
    "exp",
    "log",
    "sin",
    "sinh",
    "sqrt",
    "struct",
    "tan",
    "tanh",
]


@overload
def abs(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def abs(input: "struct.Map") -> "struct.Map":
    ...


@overload
def abs(input: "struct.Profile") -> "struct.Profile":
    ...


def abs(input: "struct.Struct") -> "struct.Struct":
    """
    Element-wise absolute value operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.abs(input.data), input.header)


@overload
def sqrt(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def sqrt(input: "struct.Map") -> "struct.Map":
    ...


@overload
def sqrt(input: "struct.Profile") -> "struct.Profile":
    ...


def sqrt(input: "struct.Struct") -> "struct.Struct":
    """
    Element-wise square root operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.sqrt(input.data), input.header)


@overload
def cbrt(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def cbrt(input: "struct.Map") -> "struct.Map":
    ...


@overload
def cbrt(input: "struct.Profile") -> "struct.Profile":
    ...


def cbrt(input: "struct.Struct") -> "struct.Struct":
    """
    Element-wise cube root operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.cbrt(input.data), input.header)


@overload
def exp(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def exp(input: "struct.Map") -> "struct.Map":
    ...


@overload
def exp(input: "struct.Profile") -> "struct.Profile":
    ...


def exp(input: "struct.Struct"):
    """
    Element-wise exponential operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.exp(input.data), input.header)


@overload
def log(input: "struct.Cube", base: Optional[float] = None) -> "struct.Cube":
    ...


@overload
def log(input: "struct.Map", base: Optional[float] = None) -> "struct.Map":
    ...


@overload
def log(input: "struct.Profile", base: Optional[float] = None) -> "struct.Profile":
    ...


def log(input: "struct.Struct", base: Optional[float] = None):
    """
    Element-wise logarithm operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.
    base : `float | None`, optional
        Base of the logarithm (by default natural logarithm). Must a positive number.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    if base is None:
        return type(input)(np.log(input.data), input.header)
    return type(input)(np.log(input.data) / np.log(base), input.header)


@overload
def cos(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def cos(input: "struct.Map") -> "struct.Map":
    ...


@overload
def cos(input: "struct.Profile") -> "struct.Profile":
    ...


def cos(input: "struct.Struct"):
    """
    Element-wise cosine operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.cos(input.data), input.header)


@overload
def sin(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def sin(input: "struct.Map") -> "struct.Map":
    ...


@overload
def sin(input: "struct.Profile") -> "struct.Profile":
    ...


def sin(input: "struct.Struct"):
    """
    Element-wise sine operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.sin(input.data), input.header)


@overload
def tan(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def tan(input: "struct.Map") -> "struct.Map":
    ...


@overload
def tan(input: "struct.Profile") -> "struct.Profile":
    ...


def tan(input: "struct.Struct"):
    """
    Element-wise tangent operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.tan(input.data), input.header)


@overload
def arccos(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def arccos(input: "struct.Map") -> "struct.Map":
    ...


@overload
def arccos(input: "struct.Profile") -> "struct.Profile":
    ...


def arccos(input: "struct.Struct"):
    """
    Element-wise inverse cosine operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.arccos(input.data), input.header)


@overload
def arcsin(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def arcsin(input: "struct.Map") -> "struct.Map":
    ...


@overload
def arcsin(input: "struct.Profile") -> "struct.Profile":
    ...


def arcsin(input: "struct.Struct"):
    """
    Element-wise inverse sine operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.arcsin(input.data), input.header)


@overload
def arctan(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def arctan(input: "struct.Map") -> "struct.Map":
    ...


@overload
def arctan(input: "struct.Profile") -> "struct.Profile":
    ...


def arctan(input: "struct.Struct"):
    """
    Element-wise inverse tangent operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.arctan(input.data), input.header)


@overload
def cosh(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def cosh(input: "struct.Map") -> "struct.Map":
    ...


@overload
def cosh(input: "struct.Profile") -> "struct.Profile":
    ...


def cosh(input: "struct.Struct"):
    """
    Element-wise hyperbolic cosine operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.cosh(input.data), input.header)


@overload
def sinh(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def sinh(input: "struct.Map") -> "struct.Map":
    ...


@overload
def sinh(input: "struct.Profile") -> "struct.Profile":
    ...


def sinh(input: "struct.Struct"):
    """
    Element-wise hyperbolic sine operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.sinh(input.data), input.header)


@overload
def tanh(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def tanh(input: "struct.Map") -> "struct.Map":
    ...


@overload
def tanh(input: "struct.Profile") -> "struct.Profile":
    ...


def tanh(input: "struct.Struct"):
    """
    Element-wise hyperbolic tangent operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.tanh(input.data), input.header)


@overload
def arccosh(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def arccosh(input: "struct.Map") -> "struct.Map":
    ...


@overload
def arccosh(input: "struct.Profile") -> "struct.Profile":
    ...


def arccosh(input: "struct.Struct"):
    """
    Element-wise inverse hyperbolic cosine operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.arccosh(input.data), input.header)


@overload
def arcsinh(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def arcsinh(input: "struct.Map") -> "struct.Map":
    ...


@overload
def arcsinh(input: "struct.Profile") -> "struct.Profile":
    ...


def arcsinh(input: "struct.Struct"):
    """
    Element-wise inverse hyperbolic sine operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.arcsinh(input.data), input.header)


@overload
def arctanh(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def arctanh(input: "struct.Map") -> "struct.Map":
    ...


@overload
def arctanh(input: "struct.Profile") -> "struct.Profile":
    ...


def arctanh(input: "struct.Struct"):
    """
    Element-wise inverse hyperbolic tangent operator.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.arctanh(input.data), input.header)



import pickle

from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import numpy as np

if __name__ == "__main__":

    values = [i/8 for i in range(8)]
    colors = [(0, 0, 0), (0, 0, 1.0), (0, 1.0, 1.0), (0, 1.0, 0), (1.0, 1.0, 0), (1.0, 0, 0), (1.0, 0, 1.0), (1.0, 1.0, 1.0)]
    norm = plt.Normalize(min(values), max(values))
    cmap = LinearSegmentedColormap.from_list(
        '', [(norm(value), tuple(np.array(color))) for value, color in zip(values, colors)])

    with open("astro_cmap.pickle", "wb") as f:

        pickle.dump(cmap, f)



import warnings
from typing import List, Optional, Union, overload

import numpy as np

from .. import structures as struct
from ..core import header as hdr

__all__ = ["integral", "spectrum", "noise_map", "reduce_spectral", "reduce_spatial"]


def integral(
    cube: "struct.Cube",
    ppv_mask: Optional["struct.Cube"] = None,
    ignore_nans: bool = True,
) -> "struct.Map":
    """
    Returns the integral of `cube` over it spectral axis. Output unit is [cube unit] * km/s.

    Parameters
    ----------

    cube : `Cube`
        Input line cube.

    Returns
    -------

    integ : `Map`
        Output map such that `integ.nx == cube.nx` and `integ.ny == cube.ny`.

    """
    if ppv_mask is not None:
        cube = ppv_mask * cube
    return (
        cube.sum(struct.Map, ignore_nans=ignore_nans) * cube.header["CDELT3"] / 1e3
    )  # Convert to km/s


def spectrum(cube: "struct.Cube") -> "struct.Profile":
    """
    Returns the mean spectrum of `cube` ,i.e., the channel-wise average.

    Parameters
    ----------

    cube : `Cube`
        Input line cube.

    Returns
    -------

    spectrum : `Profile`
        Output spectrum such that `spectrum.nz == cube.nz`.

    """
    new_data = np.nanmean(cube.data, axis=(1, 2))
    new_header = hdr.remove_header_axis(cube.header, "spatial")
    return struct.Profile(new_data, new_header)


# TODO: noise_map -> permettre aussi d'utiliser un masque 3D
def noise_map(
    cube: "struct.Cube", signal_mask: Union[slice, List[slice]], unit="index"
) -> "struct.Map":
    """
    Returns the noise map (pixel-wise standard deviation of noise) of `cube` by hiding
    the velocity channels containing signal. If every channel contains signal,
    the noise map will be filled with `NaNs`.

    Parameters
    ----------

    cube : `Cube`
        Input line cube.
    signal_mask : `slice | list[slice]`
        Intervals of channels to hide. Only step of 1 is supported.
    unit : `str`, optional
        Describe how to read the bound values of signal_mask. Must be `'index'`, `'velocity'` or `'frequency'`.
        If `unit == 'index'`, the values are numpy indexes (starting from 0).
        If `unit == 'velocity'`, the values are in km/s.
        If `unit == 'frequency'`, the values are in GHz.

    Returns
    -------

    noise_map : `Map`
        Noise map of `cube`.
    """
    unit = unit.lower()
    if isinstance(signal_mask, slice):
        signal_mask = [signal_mask]

    new_data = cube.data.copy()
    for mask in signal_mask:
        if mask.step is not None and mask.step != 1:
            raise ValueError("Slice step different of 1 is not supported")
        if unit == "index":
            index_mask = mask
        elif unit == "velocity":
            index_mask = slice(
                hdr.velocity_to_index(cube.header, mask.start),
                hdr.velocity_to_index(cube.header, mask.stop),
            )
        elif unit == "frequency":  # Order is reversed for frequencies
            index_mask = slice(
                hdr.frequency_to_index(cube.header, mask.stop) - 1,
                hdr.frequency_to_index(cube.header, mask.start) + 1,
            )
        else:
            raise ValueError(
                f"unit must be 'index', 'velocity' or 'frequency', not '{unit}'"
            )
        new_data[index_mask] *= np.nan

    with warnings.catch_warnings():  # Suppress warning when a profile is full of NaNs
        warnings.simplefilter("ignore")
        new_data = np.nanstd(new_data, axis=0)
    new_header = hdr.remove_header_axis(cube.header, "spectral")
    return struct.Map(new_data, new_header)


@overload
def reduce_spectral(
    input: "struct.Cube", z_interv: Optional[slice] = None
) -> "struct.Cube":
    ...


@overload
def reduce_spectral(
    input: "struct.Profile", z_interv: Optional[slice] = None
) -> "struct.Profile":
    ...


def reduce_spectral(
    input: Union["struct.Cube", "struct.Profile"],
    z_interv: Optional[slice] = None,
    unit: str = "index",
) -> Union["struct.Cube", "struct.Profile"]:
    """
    z_interv must be a slice (indices begin to zero)
    unit must be 'index', 'velocity' or 'frequency'
    """
    unit = unit.lower()
    if unit not in ["index", "velocity", "frequency"]:
        raise ValueError(
            f"unit must be 'index', 'velocity' or 'frequency', not '{unit}'"
        )

    # Handle Nones
    if z_interv is None:
        return input.copy()

    # Handle unit
    if unit == "index":
        z = [z_interv.start or 0, (z_interv.stop or input.nz) - 1]
    elif unit == "velocity":
        z = [
            hdr.velocity_to_index(z_interv.start) if z_interv.start is not None else 0,
            (
                hdr.velocity_to_index(z_interv.stop)
                if z_interv.stop is not None
                else input.nz
            )
            - 1,
        ]
    elif unit == "frequency":  # Order is reversed for 'frequency' unit
        z = [
            (
                hdr.frequency_to_index(z_interv.stop)
                if z_interv.stop is not None
                else input.nz
            )
            - 1,
            hdr.frequency_to_index(z_interv.start) if z_interv.start is not None else 0,
        ]

    new_data = input.data[z_interv]
    new_header = hdr.change_coordinates(input.header, z=z)
    return type(input)(new_data, new_header)


@overload
def reduce_spatial(
    input: "struct.Cube",
    x_interv: Optional[slice] = None,
    y_interv: Optional[slice] = None,
) -> "struct.Cube":
    ...


@overload
def reduce_spatial(
    input: "struct.Map",
    x_interv: Optional[slice] = None,
    y_interv: Optional[slice] = None,
) -> "struct.Map":
    ...


def reduce_spatial(
    input: Union["struct.Cube", "struct.Map"],
    x_interv: Optional[slice] = None,
    y_interv: Optional[slice] = None,
    unit: str = "index",
) -> Union["struct.Cube", "struct.Map"]:
    """
    x_interv must be a slice (indices begin to zero)
    y_interv must be a slice (indices begin to zero)
    unit must be 'index' or 'angle'
    """
    unit = unit.lower()
    if unit not in ["index", "angle"]:
        raise ValueError(f"unit must be 'index', or 'angle', not '{unit}'")

    # Handle Nones
    if x_interv is None:
        x_interv = slice(None, None)
    if y_interv is None:
        y_interv = slice(None, None)

    # Handle unit (TODO: gérer les None -> pour le moment incorrect)
    if unit == "index":
        x = [x_interv.start, x_interv.stop - 1]
        y = [y_interv.start, y_interv.stop - 1]
    elif unit == "angle":
        x_start, y_start = hdr.coordinates_to_indices(
            input.header, x_interv.start, y_interv.start
        )
        x_stop, y_stop = hdr.coordinates_to_indices(
            input.header, x_interv.stop, y_interv.stop
        )
        x = [x_start, x_stop - 1]
        y = [y_start, y_stop - 1]

    new_data = input.data[..., y_interv, x_interv]
    new_header = hdr.change_coordinates(input.header, x=x, y=y)
    return type(input)(new_data, new_header)



""" ranch module """

from .structures import *

from .core import *
from .filtering import *
from .io import *
from .reduction import *
from .models import *
from .learning import *

__all__ = []
__all__.extend(structures.__all__)
__all__.extend(core.__all__)
__all__.extend(filtering.__all__)
__all__.extend(io.__all__)
__all__.extend(reduction.__all__)
__all__.extend(models.__all__)
__all__.extend(learning.__all__)



from typing import Callable, Literal, Optional, Sequence, Type, Union, overload

import numpy as np
from astropy.io import fits

from .. import structures as struct
from ..io import io
from . import header as hdr

__all__ = [
    "apply_element_wise",
    "astype",
    "change_axes_order",
    "clip",
    "copy",
    "cube_from",
    "cube_from_maps",
    "cube_from_profiles",
    "from_fits",
    "from_numpy",
    "is_logical",
    "isfinite",
    "isnan",
    "map_from",
    "nan_to_num",
    "ones",
    "ones_like",
    "profile_from",
    "stack_numpy",
    "to_numpy",
    "where",
    "x_axis",
    "y_axis",
    "z_axis",
    "zeros",
    "zeros_like",
]

# Axes handler


@overload
def change_axes_order(
    input: "struct.Cube", order: Literal["xyz", "yxz", "zxy", "zyx"]
) -> "struct.Cube":
    ...


@overload
def change_axes_order(input: "struct.Map", order: Literal["xy", "yx"]) -> "struct.Map":
    ...


def change_axes_order(
    input: Union["struct.Cube", "struct.Map"], order: str
) -> Union["struct.Cube", "struct.Map"]:
    """TODO"""
    header = input.header
    data = input.data

    new_header = hdr.move_header_axes(header, order)

    axes_sources = {
        2: (0, 1),
        3: (0, 1, 2),
    }

    axes_destinations = {
        "xyz": (0, 1, 2),
        "yxz": (0, 2, 1),
        "zxy": (1, 2, 0),
        "zyx": (2, 1, 0),
        "xy": (0, 1),
        "yx": (1, 0),
    }

    new_data = np.moveaxis(
        data, axes_sources[len(order)], axes_destinations[order.strip().lower()]
    )

    return type(input)(new_data, new_header)


# Element wise application


@overload
def apply_element_wise(
    input: "struct.Cube", fun: Callable[[np.ndarray], np.ndarray]
) -> "struct.Cube":
    ...


@overload
def apply_element_wise(
    input: "struct.Map", fun: Callable[[np.ndarray], np.ndarray]
) -> "struct.Map":
    ...


@overload
def apply_element_wise(
    input: "struct.Profile", fun: Callable[[np.ndarray], np.ndarray]
) -> "struct.Profile":
    ...


def apply_element_wise(
    input: "struct.Struct", fun: Callable[[np.ndarray], np.ndarray]
) -> "struct.Struct":
    """
    Apply the element-wise operator `fun` on the `input` structure.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.
    fun : `Callable[[np.ndarray], ndarray]`
        Element-wise operator.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    new_data = fun(input.data)
    return type(input)(new_data, input.header)


def is_logical(input: "struct.Struct") -> bool:
    """TODO"""
    return ((input.data == 0) | (input.data == 1)).all()


# Floating type conversion


@overload
def astype(input: "struct.Cube", dtype: Literal["float", "double"]) -> "struct.Cube":
    ...


@overload
def astype(input: "struct.Map", dtype: Literal["float", "double"]) -> "struct.Map":
    ...


@overload
def astype(
    input: "struct.Profile", dtype: Literal["float", "double"]
) -> "struct.Profile":
    ...


def astype(
    input: "struct.Struct", dtype: Literal["float", "double"]
) -> "struct.Struct":
    """
    Return the input structure with floating type 'float' or 'double'.
    Note that there is no function as_int or as_float because structures are always of type float.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.
    dtype : str
        Floating type of output data. Must be 'float' or 'double'.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    if dtype.lower() == "float":
        return type(input)(input.data.astype(np.single), input.header)
    if dtype.lower() == "double":
        return type(input)(input.data.astype(np.double), input.header)
    raise ValueError("dtype must be 'float' or 'double', not {dtype}")


# Element wise test for nan values


@overload
def isnan(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def isnan(input: "struct.Map") -> "struct.Map":
    ...


@overload
def isnan(input: "struct.Profile") -> "struct.Profile":
    ...


def isnan(input: "struct.Struct") -> "struct.Struct":
    """
    Return a structure similar to `input` where a sample is 1 if input same sample is `nan` and 0 if is not.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.isnan(input.data), input.header)


@overload
def isfinite(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def isfinite(input: "struct.Map") -> "struct.Map":
    ...


@overload
def isfinite(input: "struct.Profile") -> "struct.Profile":
    ...


def isfinite(input: "struct.Struct") -> "struct.Profile":
    """
    Return a structure similar to `input` where a sample is 1 if input same sample is finite and 0 if is not.
    A finite element is a value different of `nan`, `inf` or `neginf`.
    This function is the opposite to isnan because `inf` and `neginf` are automatically casted to `nan`
    in structures constructors. So in practice : isnan(input) == ~isfinite(input).

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.isfinite(input.data), input.header)


# Nan functions


@overload
def nan_to_num(input: "struct.Cube", value: float = 0.0) -> "struct.Cube":
    ...


@overload
def nan_to_num(input: "struct.Map", value: float = 0.0) -> "struct.Map":
    ...


@overload
def nan_to_num(input: "struct.Profile", value: float = 0.0) -> "struct.Profile":
    ...


def nan_to_num(input: "struct.Struct", value: float = 0.0) -> "struct.Struct":
    """
    Returns a copy of `input` where the nans have been replaced by `value`.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.
    value : float, optional
        Value to replace `nans` elements. Default : 0.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(np.nan_to_num(input.data, nan=value), input.header)


# Initializers


@overload
def copy(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def copy(input: "struct.Map") -> "struct.Map":
    ...


@overload
def copy(input: "struct.Profile") -> "struct.Profile":
    ...


def copy(input: "struct.Struct") -> "struct.Struct":
    """
    Returns a copy of `input`.

    Parameters
    ----------
    input : `Cube | Map | Profile`
        Input structure.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    return type(input)(input.data.copy(), input.header.copy())


@overload
def from_numpy(
    cls: Type["struct.Cube"],
    array: np.ndarray,
    header: fits.Header,
    axes: Optional[Literal["xyz", "yxz", "zxy", "zyx"]] = None,
) -> "struct.Cube":
    ...


@overload
def from_numpy(
    cls: Type["struct.Map"],
    array: np.ndarray,
    header: fits.Header,
    axes: Optional[Literal["xy", "yx"]] = None,
) -> "struct.Map":
    ...


@overload
def from_numpy(
    cls: Type["struct.Profile"],
    array: np.ndarray,
    header: fits.Header,
    axes: None = None,
) -> "struct.Profile":
    ...


def from_numpy(
    cls: Type["struct.Struct"],
    array: np.ndarray,
    header: fits.Header,
    axes: Optional[str] = None,
) -> "struct.Struct":
    """
    Create an inputect of type `cls` from numpy array `array` and astropy header `header`.

    Parameters
    ----------
    cls : `Type[Cube] | Type[Map] | Type[Profile]`
        Type of structure to create.
    array : np.ndarray
        Data of structure to create. Must be 1D if cls is `Profile`, 2D if cls is `Map` and
        3D if cls is `Cube`.
    header : fits.Header
        Astropy fits header. Can be taken from another cube or created by hand.
    axes : `str`, optional
        Order of axes. Must be 'xyz', 'yxz', 'zxy', 'zyx', 'xy', 'yx' or None.

    Returns
    -------
    out : `Cube | Map | Profile`
        Created structure.
    """
    if axes is not None:
        header = hdr.move_header_axes(header, axes)
    return cls(array, header)


@overload
def from_fits(
    cls: Type["struct.Cube"],
    filename: str,
    path: str = None,
    axes=Optional[Literal["xyz", "yxz", "zxy", "zyx"]],
) -> "struct.Cube":
    ...


@overload
def from_fits(
    cls: Type["struct.Map"],
    filename: str,
    path: str = None,
    axes: Optional[Literal["xy", "yx"]] = None,
) -> "struct.Map":
    ...


@overload
def from_fits(
    cls: Type["struct.Profile"], filename: str, path: str = None, axes: None = None
) -> "struct.Profile":
    ...


def from_fits(
    cls: Type["struct.Struct"],
    filename: str,
    path: str = None,
    axes: Optional[str] = None,
) -> "struct.Struct":
    """
    Load an inputect of type `cls` from file `filename/path`.

    Parameters
    ----------
    cls : `Type[Cube] | Type[Map] | Type[Profile]`
        Type of structure to create.
    filename : `str`
        Filename of FITS file to load. Extension can be ommited. Handle both .fits or .fits.gz files,
        but notice that .gz files take longer to load.
    path : `str`, optional
        Path to the FITS file.
    axes : `str`, optional
        Order of axes. Must be 'xyz', 'yxz', 'zxy', 'zyx', 'xy', 'yx' or None.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    data, header = io._load_fits(filename, path)
    if axes is not None:
        header = hdr.move_header_axes(header, axes)
    return cls(data.astype(np.float32), header)


@overload
def zeros(cls: Type["struct.Cube"], header: fits.Header) -> "struct.Cube":
    ...


@overload
def zeros(cls: Type["struct.Map"], header: fits.Header) -> "struct.Map":
    ...


@overload
def zeros(cls: Type["struct.Profile"], header: fits.Header) -> "struct.Profile":
    ...


def zeros(cls: Type["struct.Struct"], header: fits.Header) -> "struct.Struct":
    """
    Create an inputect of type `cls` fill with zeros from astropy header `header`.

    Parameters
    ----------
    cls : `Type[Cube] | Type[Map] | Type[Profile]`
        Type of structure to create.
    header : fits.Header
        Astropy fits header. Can be taken from another cube or created by hand.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    shape = [header[f"NAXIS{k}"] for k in range(header["NAXIS"], 0, -1)]
    return from_numpy(cls, np.zeros(shape), header)


@overload
def ones(cls: Type["struct.Cube"], header: fits.Header) -> "struct.Cube":
    ...


@overload
def ones(cls: Type["struct.Map"], header: fits.Header) -> "struct.Map":
    ...


@overload
def ones(cls: Type["struct.Profile"], header: fits.Header) -> "struct.Profile":
    ...


def ones(cls: Type["struct.Struct"], header: fits.Header) -> "struct.Struct":
    """
    Create an inputect of type `cls` fill with ones from astropy header `header`.

    Parameters
    ----------
    cls : `Type[Cube] | Type[Map] | Type[Profile]`
        Type of structure to create.
    header : fits.Header
        Astropy fits header. Can be taken from another cube or created by hand.

    Returns
    -------
    out : `Cube | Map | Profile`
        Output structure.
    """
    shape = [header[f"NAXIS{k}"] for k in range(header["NAXIS"], 0, -1)]
    print(shape)
    return from_numpy(cls, np.ones(shape), header)


@overload
def zeros_like(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def zeros_like(input: "struct.Map") -> "struct.Map":
    ...


@overload
def zeros_like(input: "struct.Profile") -> "struct.Profile":
    ...


def zeros_like(input: "struct.Struct"):
    """
    Returns a structure with the same shape than input filled with zeros.

    Parameters
    ----------
    input : Cube | Map | Profile
        Input structure

    Returns
    -------
    out : Cube | Map | Profile
        Output structure
    """
    return type(input)(np.zeros_like(input.data), input.header)


@overload
def ones_like(input: "struct.Cube") -> "struct.Cube":
    ...


@overload
def ones_like(input: "struct.Map") -> "struct.Map":
    ...


@overload
def ones_like(input: "struct.Profile") -> "struct.Profile":
    ...


def ones_like(input: "struct.Struct") -> "struct.Struct":
    """
    Returns a structure with the same shape than input filled with ones.

    Parameters
    ----------
    input : Cube | Map | Profile
        Input structure

    Returns
    -------
    out : Cube | Map | Profile
        Output structure
    """
    return type(input)(np.ones_like(input.data), input.header)


# Derived initializers


def map_from(cube: "struct.Cube", value: float = 0.0):
    """
    Returns a map with the same x and y axis than `cube` filled with `value`.

    Parameters
    ----------
    cube : Cube
        Input cube.

    Returns
    -------
    out : Map
        Output map.
    """
    new_header = hdr.remove_header_axis(cube.header, axis="spectral")
    new_data = np.zeros((cube.ny, cube.nx)) + value
    return struct.Map(new_data, new_header)


def profile_from(cube: "struct.Cube", value: float = 0.0):
    """
    Returns a profile with the same z axis than `cube` filled with `value`.

    Parameters
    ----------
    cube : Cube
        Input cube.

    Returns
    -------
    out : Profile
        Output profile.
    """
    new_header = hdr.remove_header_axis(cube.header, axis="spatial")
    new_data = np.zeros(cube.nz) + value
    return struct.Profile(new_data, new_header)


def cube_from(map: "struct.Map", profile: "struct.Profile", value: float = 0.0):
    """
    Returns a cube with the same x and y axis than `map` and the same z axis
    than profile filled with `value`.

    Parameters
    ----------
    map : Map
        Input map.
    profile : Profile
        Input profile.

    Returns
    -------
    cube : Cube
        Output cube.
    """
    new_header = hdr.merge_headers(map.header, profile.header)
    new_data = np.zeros(profile.shape + map.shape) + value
    return struct.Cube(new_data, new_header)


def cube_from_maps(maps: Sequence["struct.Map"]) -> "struct.Cube":
    """
    Returns a cube builded by concatening maps in `maps` sequences.

    Parameters
    ----------
    maps : Sequence[Map]
        Sequence of maps of same shape. Must not be empty.

    Returns
    -------
    cube : Cube
        Output cube with same spatial shape than elements of maps.
    """
    if len(maps) == 0:
        raise ValueError("maps must not be empty")
    nz, ny, nx = len(maps), maps[0].ny, maps[0].nx
    new_data = np.zeros((nz, ny, nx))
    new_header = hdr.create_header("cube", maps[0].header, nz=nz)
    for k in range(nz):
        new_data[k, :, :] = maps[k].data
    return struct.Cube(new_data, new_header)


def cube_from_profiles(profiles: Sequence[Sequence["struct.Profile"]]) -> "struct.Cube":
    """
    Returns a cube builded by concatening maps in `maps` sequences.

    Parameters
    ----------
    profiles : Sequence[Sequence[Profile]]
        Sequence of sequence of profiles of same shape.
        The sequence must not be empty and each sub-sequence must also not be empty.
        profiles[i][j] is the pixel of the i-th row and the j-th column.

    Returns
    -------
    cube : Cube
        Output cube with same spectral shape than elements of profiles.
    """
    if len(profiles) == 0:
        raise ValueError("profiles must not be empty")
    for _, e in enumerate(profiles):
        if len(e) == 0:
            raise ValueError("Each element of profiles must be a non-empty sequence")
    ny, nx, nz = len(profiles), len(profiles[0]), profiles[0][0].nz
    new_data = np.zeros((nz, ny, nx))
    new_header = hdr.create_header("cube", profiles[0][0].header, nx=nx, ny=ny)
    for i in range(ny):
        for j in range(nx):
            new_data[:, j, i] = profiles[j][i].data
    return struct.Cube(new_data, new_header)


# Other functions


@overload
def where(
    bool_input: "struct.Cube",
    input_1: Union["struct.Cube", float],
    input_2: Union["struct.Cube", float],
) -> "struct.Cube":
    ...


@overload
def where(
    bool_input: "struct.Map",
    input_1: Union["struct.Map", float],
    input_2: Union["struct.Map", float],
) -> "struct.Map":
    ...


@overload
def where(
    bool_input: "struct.Profile",
    input_1: Union["struct.Profile", float],
    input_2: Union["struct.Profile", float],
) -> "struct.Profile":
    ...


def where(
    logical_input: "struct.Struct",
    input_1: Union["struct.Struct", float],
    input_2: Union["struct.Struct", float],
) -> "struct.Struct":
    """TODO"""
    if not isinstance(logical_input, struct.Struct):
        raise TypeError(
            f"logical_input must be an instance of Struct (Cube, Map or Profile), not {type(logical_input)}"
        )
    if not is_logical(logical_input):
        raise TypeError(
            f"logical_input must be logical i.e. contains only 0 and 1 samples"
        )
    a = input_1.data if isinstance(input_1, struct.Struct) else input_1
    b = input_2.data if isinstance(input_2, struct.Struct) else input_2
    new_data = np.where(logical_input.data == 1, a, b)
    return type(logical_input)(new_data, logical_input.header)


@overload
def clip(
    input: "struct.Cube", vmin: Optional[float], vmax: Optional[float]
) -> "struct.Cube":
    ...


@overload
def clip(
    input: "struct.Map", vmin: Optional[float], vmax: Optional[float]
) -> "struct.Map":
    ...


@overload
def clip(
    input: "struct.Profile", vmin: Optional[float], vmax: Optional[float]
) -> "struct.Profile":
    ...


def clip(
    input: "struct.Struct", vmin: Optional[float], vmax: Optional[float]
) -> "struct.Struct":
    """TODO"""
    return type(input)(np.clip(input.data, vmin, vmax), input.header)


# To numpy


def to_numpy(input: "struct.Struct", item: str) -> np.ndarray:
    """TODO"""
    if item.strip().lower() not in ["pixel", "map"]:
        raise ValueError(f"item must be 'pixel' or 'map', not {item}")
    if isinstance(input, struct.Cube):
        if item == "pixel":
            return input.data.reshape(-1, input.data.shape[1] * input.data.shape[2]).T
        else:
            return input.data
    elif isinstance(input, struct.Map):
        if item == "pixel":
            return input.data.flatten()
        else:
            raise ValueError("item = 'map' is not compatible with a Map inputect")
    elif isinstance(input, struct.Profile):
        if item == "pixel":
            raise ValueError(
                "item = 'profile' is not compatible with a Profile inputect"
            )
        else:
            return input.data
    else:
        raise TypeError(
            f"input must be an instance of Cube, Map or Profile, not {type(input)}"
        )


def stack_numpy(arrays: Sequence[np.ndarray]) -> np.ndarray:
    """TODO"""
    return np.concatenate(arrays, axis=0)


# Axes


def x_axis(
    input: Union["struct.Map", "struct.Cube"], unit: Literal["index", "angle"] = "index"
) -> np.ndarray:
    """TODO"""
    if not isinstance(input, (struct.Map, struct.Cube)):
        raise TypeError(f"input must be an instance of Map or Cube, not {type(input)}")
    unit = unit.strip().lower()
    if unit not in ["index", "angle"]:
        raise ValueError(f"unit must be 'index' or 'angle', not '{unit}'")

    axis = np.arange(input.nx)
    if unit == "angle":
        axis = hdr.indices_to_coordinates(input.header, axis, 0)[0]
    return axis


def y_axis(
    input: Union["struct.Map", "struct.Cube"], unit: Literal["index", "angle"] = "index"
) -> np.ndarray:
    """TODO"""
    if not isinstance(input, (struct.Map, struct.Cube)):
        raise TypeError(f"input must be an instance of Map or Cube, not {type(input)}")
    unit = unit.strip().lower()
    if unit not in ["index", "angle"]:
        raise ValueError(f"unit must be 'index' or 'angle', not '{unit}'")

    axis = np.arange(input.ny)
    if unit == "angle":
        axis = hdr.indices_to_coordinates(input.header, axis, 0)[1]
    return axis


def z_axis(
    input: Union["struct.Profile", "struct.Cube"],
    unit: Literal["index", "velocity", "frequency"] = "index",
) -> np.ndarray:
    """TODO"""
    if not isinstance(input, (struct.Profile, struct.Cube)):
        raise TypeError(
            f"input must be an instance of Profile or Cube, not {type(input)}"
        )
    unit = unit.strip().lower()
    if unit not in ["index", "velocity", "frequency"]:
        raise ValueError(
            f"unit must be 'index', 'velocity' or 'frequency', not '{unit}'"
        )

    axis = np.arange(input.nz)
    if unit == "velocity":
        axis = hdr.index_to_velocity(input.header, axis)
    elif unit == "frequency":
        axis = hdr.index_to_frequency(input.header, axis)
    return axis



from math import sqrt, comb
from typing import List, Optional
from warnings import warn

from scipy.special import erfinv

from .. import structures as struct
from . import Kernels

__all__ = [
    'noise_clipping'
]

def noise_clipping(cube : 'struct.Cube', percent : float,
    noise_map : Optional['struct.Map'] = None, signal_channels : Optional[List[slice]] = None,
    p_xy : int = 5, p_z : int = 3, type_xy : str = 'gaussian', type_z : str = 'gaussian',
    nans : bool = False, verbose : bool = False) :

    if noise_map is not None and signal_channels is not None :
        warn('noise_map and signal_channels are both not None so signal_channels will be ignored')
    elif noise_map is None and signal_channels is not None :
        noise_map = cube.noise_map(signal_channels)
    elif signal_channels is None and noise_map is not None :
        pass
    else :
        raise ValueError('noise_map and signal_channels must not be both None')

    # Spatial and spectral clipping
    if type_xy.lower() == 'gaussian' :
        h_xy = Kernels.gaussian_1d(p_xy)
        sigma_xy = 1/2**(p_xy-1)*sqrt(comb(2*(p_xy-1), p_xy-1))
    elif type_xy.lower() == 'uniform' :
        h_xy = Kernels.uniform_1d(p_xy)
        sigma_xy = sqrt(1/p_xy)

    if type_z.lower() == 'gaussian' :
        h_z = Kernels.gaussian_1d(p_z)
        sigma_z = 1/2**(p_z-1)*sqrt(comb(2*(p_z-1), p_z-1))
    elif type_z.lower() == 'uniform' :
        h_z = Kernels.uniform_1d(p_z)
        sigma_z = sqrt(1/p_z)
        
    h = Kernels.assemble_separables(h_z, h_xy, h_xy)
    
    sigma : float = sigma_xy**2 * sigma_z
    thresh : float = sqrt(2) * sigma * erfinv(2*percent/100-1)
    if verbose :
        print('Threshold : {:.2f}'.format(thresh))

    m = (cube/noise_map).filter(h)
    mask = m > thresh
    cube_clipped = mask.where(cube, 0. if not nans else float('nan'))

    return cube_clipped, mask



from typing import Union, overload

from scipy.ndimage import maximum_filter1d

from .. import structures as struct
from ..core import header as hdr

__all__ = ["peaks_detection"]


@overload
def peaks_detection(
    input: "struct.Cube",
    delta_peaks: int = 1,
    peaks_min: Union[float, "struct.Struct"] = -float("inf"),
) -> "struct.Cube":
    ...


@overload
def peaks_detection(
    input: "struct.Profile",
    delta_peaks: int = 1,
    peaks_min: Union[float, "struct.Struct"] = -float("inf"),
) -> "struct.Profile":
    ...


def peaks_detection(
    input: Union["struct.Cube", "struct.Profile"],
    delta_peaks: int = 1,
    peaks_min: Union[float, "struct.Struct"] = -float("inf"),
) -> "struct.Struct":
    """axis must be 'spectral', 'spatial' or 'all'"""

    peaks = type(input).from_numpy(
        maximum_filter1d(input.data, 2 * delta_peaks + 1, axis=0), input.header
    )
    peaks = (peaks == input) & (input > peaks_min)

    return peaks



from math import comb
from typing import Callable, Union
from warnings import warn

import numpy as np
from scipy import ndimage

from .. import structures as struct

__all__ = [
    'Kernels', 'filter_channels', 'filter_cube', 'filter_map', 'filter_pixels', 'filter_profile'
]

class Kernels :
    """ Helpers to instantiate structuring elements """

    @staticmethod
    def uniform_1d(n : int) -> np.ndarray :
        """ TODO """
        return 1/n * np.ones(n)

    @staticmethod
    def gaussian_1d(n : int) -> np.ndarray :
        """ TODO """
        return 1/2**(n-1) * np.array([comb(n-1, k) for k in range(n)])

    @staticmethod
    def assemble_separables(*h) -> np.ndarray :
        """ TODO """
        H = h[0]
        for i, hi in enumerate(h[1:], 1) :
            H = np.expand_dims(H, -1) * np.expand_dims(hi, tuple(range(i)))
        return H


def _is_binary_filter(h : np.ndarray) -> bool :
    """ Returns True if h is a binary filter i.e. h contains only 0 and 1, else returns False """
    return ((h == 0) | (h == 1)).all()

def filter_pixels(cube : 'struct.Cube', kernel : np.ndarray, filtering_mode : str = 'linear',
    padding_mode : str = 'constant', cval : float = 0) :
    """ TODO """
    if kernel.ndim != 1 :
        raise ValueError(f"kernel must have 1 dimension, not {kernel.ndim}")
    if filtering_mode not in ['lin', 'linear', 'med', 'median', 'min', 'minimum', 'max', 'maximum'] :
        raise ValueError(f"filtering_mode must be 'linear', 'median', 'min' or 'max', not {filtering_mode}")
    if filtering_mode not in ['lin', 'linear'] and not _is_binary_filter(kernel) :
        warn('kernel is not a binary array. Every non zero element are considered True.')

    kernel = np.expand_dims(kernel, (1,2))
    kwargs = {'mode': padding_mode, 'cval': cval}

    if filtering_mode in ['lin', 'linear'] :
        new_data = ndimage.filters.convolve(cube.data, kernel, **kwargs)
    elif filtering_mode in ['med', 'median'] :
        new_data = ndimage.filters.median_filter(cube.data, footprint = kernel, **kwargs)
    elif filtering_mode in ['min', 'minimum'] :
        new_data = ndimage.filters.minimum_filter(cube.data, footprint = kernel, **kwargs)
    elif filtering_mode in ['max', 'maximum'] :
        new_data = ndimage.filters.maximum_filter(cube.data, footprint = kernel, **kwargs)
    else :
        raise ValueError('Should never been here')
        
    return struct.Cube(new_data, cube.header)

def filter_channels(cube : 'struct.Cube', kernel : np.ndarray, filtering_mode : str = 'linear',
    padding_mode : str = 'constant', cval : float = 0) :
    """ TODO """
    if kernel.ndim != 2 :
        raise ValueError(f"kernel must have 2 dimensions, not {kernel.ndim}")
    if filtering_mode not in ['lin', 'linear', 'med', 'median', 'min', 'minimum', 'max', 'maximum'] :
        raise ValueError(f"filtering_mode must be 'linear', 'median', 'min' or 'max', not {filtering_mode}")
    if filtering_mode not in ['lin', 'linear'] and not _is_binary_filter(kernel) :
        warn('kernel is not a binary array. Every non zero element are considered True.')

    kernel = np.expand_dims(kernel, 0)
    kwargs = {'mode': padding_mode, 'cval': cval}

    print(filtering_mode)

    if filtering_mode in ['lin', 'linear'] :
        new_data = ndimage.filters.convolve(cube.data, kernel, **kwargs)
    elif filtering_mode in ['med', 'median'] :
        new_data = ndimage.filters.median_filter(cube.data, footprint = kernel, **kwargs)
    elif filtering_mode in ['min', 'minimum'] :
        new_data = ndimage.filters.minimum_filter(cube.data, footprint = kernel, **kwargs)
    elif filtering_mode in ['max', 'maximum'] :
        new_data = ndimage.filters.maximum_filter(cube.data, footprint = kernel, **kwargs)
    else :
        raise ValueError('Should never been here')

    return struct.Cube(new_data, cube.header)

def filter_cube(cube : 'struct.Cube', kernel : np.ndarray, filtering_mode : str = 'linear',
    padding_mode : str = 'constant', cval : float = 0) :
    """ TODO """
    if kernel.ndim != 3 :
        raise ValueError(f"kernel must have 3 dimensions, not {kernel.ndim}")
    if filtering_mode not in ['lin', 'linear', 'med', 'median', 'min', 'minimum', 'max', 'maximum'] :
        raise ValueError(f"filtering_mode must be 'linear', 'median', 'min' or 'max', not {filtering_mode}")
    if filtering_mode not in ['lin', 'linear'] and not _is_binary_filter(kernel) :
        warn('kernel is not a binary array. Every non zero element are considered True.')

    kwargs = {'mode': padding_mode, 'cval': cval}

    if filtering_mode in ['lin', 'linear'] :
        new_data = ndimage.filters.convolve(cube.data, kernel, **kwargs)
    elif filtering_mode in ['med', 'median'] :
        new_data = ndimage.filters.median_filter(cube.data, footprint = kernel, **kwargs)
    elif filtering_mode in ['min', 'minimum'] :
        new_data = ndimage.filters.minimum_filter(cube.data, footprint = kernel, **kwargs)
    elif filtering_mode in ['max', 'maximum'] :
        new_data = ndimage.filters.maximum_filter(cube.data, footprint = kernel, **kwargs)
    else :
        raise ValueError('Should never been here')
    
    return struct.Cube(new_data, cube.header)

def filter_map(map : 'struct.Map', kernel : np.ndarray, filtering_mode : str = 'linear',
    padding_mode : str = 'constant', cval : float = 0) :
    """ TODO """
    if kernel.ndim != 2 :
        raise ValueError(f"kernel must have 2 dimensions, not {kernel.ndim}")
    if filtering_mode not in ['lin', 'linear', 'med', 'median', 'min', 'minimum', 'max', 'maximum'] :
        raise ValueError(f"filtering_mode must be 'linear', 'median', 'min' or 'max', not {filtering_mode}")
    if filtering_mode not in ['lin', 'linear'] and not _is_binary_filter(kernel) :
        warn('kernel is not a binary array. Every non zero element are considered True.')

    kwargs = {'mode': padding_mode, 'cval': cval}

    if filtering_mode in ['lin', 'linear'] :
        new_data = ndimage.filters.convolve(map.data, kernel, **kwargs)
    elif filtering_mode in ['med', 'median'] :
        new_data = ndimage.filters.median_filter(map.data, footprint = kernel, **kwargs)
    elif filtering_mode in ['min', 'minimum'] :
        new_data = ndimage.filters.minimum_filter(map.data, footprint = kernel, **kwargs)
    elif filtering_mode in ['max', 'maximum'] :
        new_data = ndimage.filters.maximum_filter(map.data, footprint = kernel, **kwargs)
    else :
        raise ValueError('Should never been here')
    
    return struct.Map(new_data, map.header)

def filter_profile(profile : 'struct.Profile', kernel : np.ndarray, filtering_mode : str = 'linear',
    padding_mode : str = 'constant', cval : float = 0) :
    """ TODO """
    if kernel.ndim != 1 :
        raise ValueError(f"kernel must have 1 dimension, not {kernel.ndim}")
    if filtering_mode not in ['lin', 'linear', 'med', 'median', 'min', 'minimum', 'max', 'maximum'] :
        raise ValueError(f"filtering_mode must be 'linear', 'median', 'min' or 'max', not {filtering_mode}")
    if filtering_mode not in ['lin', 'linear'] and not _is_binary_filter(kernel) :
        warn('kernel is not a binary array. Every non zero element are considered True.')

    kwargs = {'mode': padding_mode, 'cval': cval}

    if filtering_mode in ['lin', 'linear'] :
        new_data = ndimage.filters.convolve(profile.data, kernel, **kwargs)
    elif filtering_mode in ['med', 'median'] :
        new_data = ndimage.filters.median_filter(profile.data, footprint = kernel, **kwargs)
    elif filtering_mode in ['min', 'minimum'] :
        new_data = ndimage.filters.minimum_filter(profile.data, footprint = kernel, **kwargs)
    elif filtering_mode in ['max', 'maximum'] :
        new_data = ndimage.filters.maximum_filter(profile.data, footprint = kernel, **kwargs)
    else :
        raise ValueError('Should never been here')
    
    return struct.Profile(new_data, profile.header)



import os
import warnings
from itertools import product
from pickle import load as pickle_load
from typing import Optional, Tuple, Union, overload

import matplotlib.image as img
import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits
from matplotlib import colorbar, colors
from mpl_toolkits.axes_grid1 import make_axes_locatable

from .. import structures as struct
from ..core import header as hdr

__all__ = [
    "default_cmap",
    "plot_channel",
    "plot_hist",
    "plot_hist2d",
    "plot_map",
    "plot_pixel",
    "plot_profile",
    "save_channel_plot",
    "save_hist",
    "save_hist2d",
    "save_map_plot",
    "save_pixel_plot",
    "save_profile_plot",
    "show_channel",
    "show_hist",
    "show_hist2d",
    "show_map",
    "show_pixel",
    "show_profile",
]

path = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(path, "astro_cmap.pickle"), "rb") as f:
    default_cmap = pickle_load(f)


# Loading and saving


def _load_fits(
    filename: str, path: Optional[str] = None
) -> Tuple[np.ndarray, fits.Header]:
    """Load FITS file. The extension of the file can be ommited. `.fits` and `.fits.gz` extension are both checked."""
    file = filename if path is None else os.path.join(path, filename)
    file = file.replace(".fits", "").replace(".gz", "")
    if os.path.exists(file + ".fits"):
        ext = ".fits"
    elif os.path.exists(file + ".fits.gz"):
        ext = ".fits.gz"
    else:
        raise FileNotFoundError(
            "No files with a .fits or .fits.gz extension were found for the path {}".format(
                file
            )
        )
    hdu = fits.open(file + ext)[0]
    return hdu.data, hdu.header


def _save_fits(
    input: "struct.Struct", filename: str, path: Optional[str] = None, overwrite=False
) -> None:
    """Save `input` structure in a FITS file. The extension of the file can be ommited."""
    if path is not None and not os.path.exists(path):
        os.makedirs(path)
    file = filename if path is None else os.path.join(path, filename)
    file = file.replace(".fits", "").replace(".gz", "") + ".fits"
    hdu = fits.PrimaryHDU(input.data, input.header)
    hdu.writeto(file, overwrite=overwrite)


# Visualization


def _colorbar(
    mappable: img.AxesImage, label: Optional[str] = None, **kwargs
) -> colorbar.Colorbar:
    """
    Creates and returns a nice looking colorbar.

    Parameters
    ----------
    mappable : img.AxesImage
        Image (basically returned by a function like imshow).
    label : Optional[str], optional
        Colorbar label. If None, no label is added, by default None.

    Returns
    -------
    colorbar.Colorbar
        Plotted colorbar.
    """
    last_axes = plt.gca()
    ax = mappable.axes
    fig = ax.figure
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="10%", pad=0.10)
    cbar = fig.colorbar(mappable, cax=cax, label=label, **kwargs)
    plt.sca(last_axes)
    return cbar


def plot_map(
    map: "struct.Map",
    ax: Optional[plt.Axes] = None,
    label_unit: str = "angle",
    no_logical=False,
    norm: Optional[colors.Normalize] = None,
    cmap: Union[str, colors.Colormap] = default_cmap,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
):
    """Plots a map. Returns the figure axis and the colorbar."""
    if not isinstance(map, struct.Map):
        raise ValueError(f"map must be an instance of Map, not {type(map)}")
    label_unit = label_unit.lower()
    if label_unit not in ["index", "angle"]:
        raise ValueError(f"label_unit must be 'index' or 'angle', not {label_unit}")

    if cmap is None:
        cmap = default_cmap

    imshow_kwargs = {"norm": norm, "vmin": vmin, "vmax": vmax, "origin": "lower"}

    if (
        map.header["CTYPE1"].upper() == "INDICES"
        or map.header["CTYPE2"].upper() == "INDICES"
    ):
        label_unit = "index"

    if label_unit == "angle":
        bounds = hdr.bound_coordinates(map.header)
        imshow_kwargs.update(
            {
                "extent": bounds[0] + bounds[1],
                "aspect": map.shape[0]
                / map.shape[1]
                * abs((bounds[0][1] - bounds[0][0]) / (bounds[1][1] - bounds[1][0])),
            }
        )
    else:
        imshow_kwargs.update(
            {"extent": (1, map.header["NAXIS1"], 1, map.header["NAXIS2"])}
        )

    if map.is_logical() and not no_logical:
        cmap = colors.LinearSegmentedColormap.from_list(
            "binary cmap", [plt.cm.gray(0.0), plt.cm.gray(1.0)], 2
        )
        imshow_kwargs.update({"vmin": 0, "vmax": 1})
        colorbar_kwargs = {"ticks": [0.25, 0.75]}
    else:
        colorbar_kwargs = {}

    if ax is not None:
        im = ax.imshow(map.data, cmap=cmap, **imshow_kwargs)
    else:
        im = plt.imshow(map.data, cmap=cmap, **imshow_kwargs)
        ax = plt.gca()

    cbar = _colorbar(im, label=None, **colorbar_kwargs)
    if map.is_logical() and not no_logical:
        cbar.ax.set_yticklabels([0, 1])

    if label_unit == "index":
        ax.set_xlabel("$x$")
        ax.set_ylabel("$y$")
    else:
        ax.set_xlabel("$\delta x$ (')")
        ax.set_ylabel("$\delta y$ (')")

    return im, ax, cbar


def plot_profile(
    profile: "struct.Profile",
    ax: plt.Axes = None,
    label_unit: str = "velocity",
    logy: bool = False,
    linestyle: Optional[str] = "solid",
    color: Optional[str] = "k",
    linewidth: Optional[float] = 1.5,
    label: Optional[str] = None,
):
    """Plot a profile. Returns the figure axis."""
    if not isinstance(profile, struct.Profile):
        raise ValueError(f"profile must be an instance of Profile, not {type(profile)}")

    plot_kwargs = {
        "linestyle": linestyle,
        "color": color,
        "linewidth": linewidth,
        "label": label,
    }

    if logy:
        plot = ax.semilogy if ax is not None else plt.semilogy
    else:
        plot = ax.plot if ax is not None else plt.plot

    plot_label = ax.set_xlabel if ax is not None else plt.xlabel

    if profile.header["CTYPE1"].upper() == "INDICES":
        label_unit = "index"

    idx = np.arange(profile.data.size)
    if label_unit == "velocity":
        line = plot(
            hdr.index_to_velocity(profile.header, idx), profile.data, **plot_kwargs
        )
        plot_label("Velocity (km/s)")
    elif label_unit == "frequency":
        line = plot(
            hdr.index_to_frequency(profile.header, idx), profile.data, **plot_kwargs
        )
        plot_label("Frequency (GHz)")
    else:
        line = plot(idx + 1, profile.data, **plot_kwargs)
        plot_label("Channel index")

    if ax is None:
        ax = plt.gca()

    return line, ax


def plot_channel(
    cube: "struct.Cube", z: Union[int, float], unit: str = "index", **kwargs
):
    """TODO"""
    return cube.get_channel(z, unit=unit).plot(**kwargs)


def plot_pixel(
    cube: "struct.Cube", xy: tuple[int, float], unit: str = "index", **kwargs
):
    """TODO"""
    return cube.get_pixel(xy, unit=unit).plot(**kwargs)


def save_map_plot(
    map: "struct.Map", filename: str, path: Optional[str] = None, **kwargs
) -> None:
    """TODO"""
    plot_map(map, **kwargs)
    if path is None:
        plt.savefig(filename)
    else:
        plt.savefig(os.path.join(path, filename))
    plt.clf()


def save_profile_plot(
    profile: "struct.Profile", filename: str, path: Optional[str] = None, **kwargs
) -> None:
    """TODO"""
    plot_profile(profile, **kwargs)
    if path is None:
        plt.savefig(filename)
    else:
        plt.savefig(os.path.join(path, filename))
    plt.clf()


def save_channel_plot(
    cube: "struct.Cube",
    z: Union[int, float],
    filename: str,
    unit: str = "index",
    path: Optional[str] = None,
    **kwargs,
) -> None:
    """TODO"""
    map = cube.get_channel(z, unit=unit)
    save_map_plot(map, filename, path, **kwargs)


def save_pixel_plot(
    cube: "struct.Cube",
    xy: Union[tuple[int], tuple[float]],
    filename: str,
    unit: str = "index",
    path: Optional[str] = None,
    **kwargs,
) -> None:
    """TODO"""
    profile = cube.get_pixel(xy, unit=unit)
    save_profile_plot(profile, filename, path, **kwargs)


def show_map(map: "struct.Map", **kwargs) -> None:
    """TODO"""
    plot_map(map, **kwargs)
    plt.show()


def show_profile(profile: "struct.Profile", **kwargs) -> None:
    """TODO"""
    plot_profile(profile, **kwargs)
    plt.show()


def show_channel(
    cube: "struct.Cube", z: Union[int, float], unit: str = "index", **kwargs
) -> None:
    """TODO"""
    map = cube.get_channel(z, unit=unit)
    show_map(map, **kwargs)


def show_pixel(
    cube: "struct.Cube",
    xy: Union[tuple[int], tuple[float]],
    unit: str = "index",
    **kwargs,
) -> None:
    """TODO"""
    profile = cube.get_pixel(xy, unit=unit)
    show_profile(profile, **kwargs)


# Histograms


def plot_hist(
    input: "struct.Struct",
    bins: Union[int, str] = "auto",
    ax: Optional[plt.Axes] = None,
    logx: bool = False,
    logy: bool = False,
    density: bool = False,
    color: Optional[str] = None,
    xlim: Optional[tuple] = None,
    ylim: Optional[tuple] = None,
) -> Tuple[np.ndarray, np.ndarray, plt.Axes]:
    """TODO"""
    a = input.data.flatten()
    a = a[~np.isnan(a)]

    if logx:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            _a = np.log10(a)
        bins = 10 ** np.histogram_bin_edges(_a[np.isfinite(_a)], bins)
    else:
        bins = np.histogram_bin_edges(a, bins)

    if ax is None:
        ax = plt.gca()

    values, bins, _ = ax.hist(a, bins=bins, log=logy, density=density, color=color)

    if logx:
        ax.set_xscale("log")
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    return values, bins, ax


def plot_hist2d(
    input1: "struct.Struct",
    input2: "struct.Struct",
    bins: Union[int, tuple, str] = "auto",
    ax: Optional[plt.Axes] = None,
    logx: bool = False,
    logy: bool = False,
    logz: bool = False,
    density: bool = False,
    plot_identity: bool = False,
    cmap: Union[str, colors.Colormap] = default_cmap,
    scatter: bool = False,
    scatter_threshold: int = 1,
    zeros_to_nans: bool = True,
    xlim: Optional[tuple] = None,
    ylim: Optional[tuple] = None,
    vmin: float = None,
    vmax: float = None,
) -> Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray], plt.Axes]:
    """
    Plot a 2-dimensional histogram of two structures of same shape `input1` and `input2`.

    Parameters
    ----------
    input1 : `Cube | Map | Profile`.
        Structure whose data will be plotted on the x axis.
    input2 : `Cube | Map | Profile`.
        Structure whose data will be plotted on the y axis.
    bins : `int | tuple | str`, optional.
        Number of bins. If it is an integer, then the same number of bins is used for each axis.
        If it is a tuple, then we assume that the format is `(bins_x, bins_y)`.
        If bins is a string, then the number of bins is automatically computed (default: `'auto'`).
        For more information, refer to `numpy.histogram_bin_edges`.
    ax : `Axes`, optional.
        Matplotlib axis to plot the histogram. If `ax` is None, then `matplotlib.pyplot.gca()` is used.
        Default: `None`.
    logx : `bool`, optional.
        If `True`, then the x axis is plotted in log scale. Default: False.
    logy : `bool`, optional.
        If `True`, then the y axis is plotted in log scale. Default: False.
    logz : `bool`, optional.
        If `True`, then the colorbar is in log scale. Default: False.
    density : `bool`, optional.
        If `True`, then the density is plotted. Else, the standard counting of samples by bin is plotted.
        Default: `False`.
    plot_identity : `bool`, optional.
        If `True`, then the y=x line is plotted on the figure. Default: `False`.
    cmap : `str`, optional
        Default: `default_colormap`.
    scatter : `bool`, optional.
        If `True`, the samples that are less than 2 per bins are displayed individually
        (using `matplotlib.pyplot.scatter`) instead of being included in the histogram. Default: `False`.
    scatter_threshold : `int`, optional.
        The minimum number of pixel by bins necessary to be plotted as an histogram and not as a scatter plot.
        Must be positive. Default: 1.
    zeros_to_nans : `bool`, optional.
        If `True`, then the bins without samples are plotted as `nan`. Default: `True`.
    xlim : `tuple | None`, optional.
        X axis limits for `matplotlib.pyplot.xlim`. Default: `None`.
    ylim : `tuple | None`, optional.
        Y axis limits for `matplotlib.pyplot.ylim`. Default: `None`.
    vmin : `float`, optional.
        Minimum value for colormap. Default: `None`.
    vmax : `float`, optional.
        Maximum value for colormap. Default: `None`.

    Returns
    -------
    H : `ndarray`
        Matrix of 2-dimensional histogram.
    edges : `tuple` of `ndarray`
        Tuple of x and y axis bins `(xedges, yedges)`
    ax : `Axes`
        Matplotlib axis.
    """
    if input1.shape != input2.shape:
        raise ValueError(
            f"input1 and input2 must have the same shape (here {input1.shape} and {input2.shape})"
        )

    a = input1.data.flatten()
    b = input2.data.flatten()

    indices = ~np.isnan(a) & ~np.isnan(b)
    a, b = a[indices], b[indices]

    if xlim is None:
        _xlim = (a.min(), a.max())
    else:
        _xlim = (
            a.min() if xlim[0] is None else xlim[0],
            a.max() if xlim[1] is None else xlim[1],
        )
    if ylim is None:
        _ylim = (b.min(), b.max())
    else:
        _ylim = (
            b.min() if ylim[0] is None else ylim[0],
            b.max() if ylim[1] is None else ylim[1],
        )

    indices = (a >= _xlim[0]) & (a <= _xlim[1]) & (b >= _ylim[0]) & (b <= _ylim[1])
    a, b = a[indices], b[indices]

    if isinstance(bins, str):
        if logx:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=RuntimeWarning)
                _a = np.log10(a)
        else:
            _a = a
        if logy:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=RuntimeWarning)
                _b = np.log10(b)
        else:
            _b = b
        _binsx = np.histogram_bin_edges(_a[np.isfinite(_a)], bins).size - 1
        _binsy = np.histogram_bin_edges(_b[np.isfinite(_b)], bins).size - 1
        _bins = round((_binsx * _binsy) ** 0.25)
        bins = (_bins, _bins)
    elif isinstance(bins, int):
        if bins <= 0:
            raise ValueError(
                "bins must be a positive integer, a tuple of two positive integers or a string"
            )
        bins = (bins, bins)
    elif isinstance(bins, tuple):
        if len(bins) != 2 or bins[0] <= 0 or bins[1] <= 0:
            raise ValueError(
                "bins must be a positive integer, a tuple of two positive integers or a string"
            )
    else:
        raise TypeError(
            f"bins must be an integer, a tuple of integer or a string, not {type(bins)}"
        )

    if logx:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            _a = np.log10(a)
        bin_edges_a = 10 ** np.histogram_bin_edges(_a[np.isfinite(_a)], bins[0])
    else:
        bin_edges_a = np.histogram_bin_edges(a, bins[0])

    if logy:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            _b = np.log10(b)
        bin_edges_b = 10 ** np.histogram_bin_edges(_b[np.isfinite(_b)], bins[1])
    else:
        bin_edges_b = np.histogram_bin_edges(b, bins[1])

    bin_edges = (bin_edges_a, bin_edges_b)

    H, xedges, yedges = np.histogram2d(a, b, bins=bin_edges, density=density)
    X, Y = np.meshgrid(xedges, yedges)

    if scatter or zeros_to_nans:
        H_count, _, __ = np.histogram2d(a, b, bins=bin_edges, density=False)

    if scatter:
        a_scatter, b_scatter = [], []
        for i, j in product(range(xedges.size - 1), range(yedges.size - 1)):
            if (0 < H_count[i, j]) and (H_count[i, j] <= scatter_threshold):
                _cond = (
                    (xedges[i] <= a)
                    & (a <= xedges[i + 1])
                    & (yedges[j] <= b)
                    & (b <= yedges[j + 1])
                )
                a_scatter += a[_cond].tolist()
                b_scatter += b[_cond].tolist()
                H[i, j] = 0

    if zeros_to_nans:
        H[H_count == 0] *= np.nan

    if vmin is not None:
        H[H < vmin] *= np.nan
    if vmax is not None:
        H[H > vmax] *= np.nan

    if ax is None:
        ax = plt.gca()

    im = ax.pcolormesh(
        X, Y, H.T, norm=colors.LogNorm(vmin, vmax) if logz else None, cmap=cmap
    )
    ax.get_figure().colorbar(im, ax=ax)

    if scatter:
        print("len", a_scatter.__len__(), b_scatter.__len__())
        ax.scatter(a_scatter, b_scatter, s=5, color="k")

    if plot_identity:
        _start = max(xedges[0], yedges[0])
        _end = min(xedges[-1], yedges[-1])
        ax.plot([_start, _end], [_start, _end], "--k")

    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)
    if logx:
        ax.set_xscale("log")
    if logy:
        ax.set_yscale("log")

    return H.T, (xedges, yedges), ax


def save_hist(
    input: "struct.Struct", filename: str, path: Optional[str] = None, **kwargs
):
    """TODO"""
    plot_hist(input, **kwargs)
    plt.savefig(os.path.join(path, filename))


def save_hist2d(
    input1: "struct.Struct",
    input2: "struct.Struct",
    filename: str,
    path: Optional[str] = None,
    **kwargs,
):
    """TODO"""
    plot_hist2d(input1, input2, **kwargs)
    plt.savefig(os.path.join(path, filename))


def show_hist(input: "struct.Struct", **kwargs):
    """TODO"""
    plot_hist(input, **kwargs)
    plt.show()


def show_hist2d(input1: "struct.Struct", input2: "struct.Struct", **kwargs):
    """TODO"""
    plot_hist2d(input1, input2, **kwargs)
    plt.show()
